from ednaml.datareaders import DataReader
from ednaml.generators.AlbertGenerator import AlbertGenerator, AlbertDataset
from torch.utils.data import Dataset
from torch.utils.data import DataLoader


class AlbertReader(DataReader):
    name: str = "AlbertReader"
    DATASET: Dataset = AlbertDataset
    GENERATOR = AlbertGenerator



import torch
import math


def gelu(x):
    """Original Implementation of the gelu activation function in Google Bert repo when initially created.
    For information: OpenAI GPT's gelu is slightly different (and gives slightly different results):
    0.5 * x * (1 + torch.tanh(math.sqrt(2 / math.pi) * (x + 0.044715 * torch.pow(x, 3))))
    Also see https://arxiv.org/abs/1606.08415
    """
    return x * 0.5 * (1.0 + torch.erf(x / math.sqrt(2.0)))


def gelu_new(x):
    """Implementation of the gelu activation function currently in Google Bert repo (identical to OpenAI GPT).
    Also see https://arxiv.org/abs/1606.08415
    """
    return (
        0.5
        * x
        * (
            1
            + torch.tanh(
                math.sqrt(2 / math.pi) * (x + 0.044715 * torch.pow(x, 3))
            )
        )
    )


def swish(x):
    return x * torch.sigmoid(x)



import importlib
import os, shutil, logging, glob, re
from types import MethodType
from typing import Callable, Dict, List, Type, Union
from urllib.error import URLError
import warnings
from torchinfo import ModelStatistics
from ednaml.config.EdnaMLConfig import EdnaMLConfig
from ednaml.config.LossConfig import LossConfig
from ednaml.config.ModelConfig import ModelConfig
from ednaml.core import EdnaMLBase
from ednaml.crawlers import Crawler
from ednaml.datareaders import DataReader
from ednaml.generators import Generator
from ednaml.loss.builders import LossBuilder
from ednaml.models.ModelAbstract import ModelAbstract
from ednaml.optimizer import BaseOptimizer
from ednaml.optimizer.StandardLossOptimizer import StandardLossOptimizer
from ednaml.loss.builders import ClassificationLossBuilder
from ednaml.plugins.ModelPlugin import ModelPlugin
from ednaml.trainer.BaseTrainer import BaseTrainer
from ednaml.storage import BaseStorage
from ednaml.utils import locate_class, path_import
import ednaml.utils
import torch
from torchinfo import summary
from ednaml.utils.LabelMetadata import LabelMetadata
import ednaml.utils.web
from ednaml.utils.SaveMetadata import SaveMetadata
import logging
import ednaml.core.decorators

"""TODO

Verbosity = 0 -> create no logger and use a dummy that print nothing anywhere
"""


class EdnaML(EdnaMLBase):
    labelMetadata: LabelMetadata
    modelStatistics: ModelStatistics
    model: ModelAbstract = None
    loss_function_array: List[LossBuilder]
    loss_optimizer_array: List[torch.optim.Optimizer]
    optimizer: List[torch.optim.Optimizer]
    scheduler: List[torch.optim.lr_scheduler._LRScheduler]
    loss_scheduler: List[torch.optim.lr_scheduler._LRScheduler]
    previous_stop: int
    trainer: BaseTrainer
    crawler: Crawler
    train_generator: Generator
    test_generator: Generator
    cfg: EdnaMLConfig
    decorator_reference: Dict[str,Type[MethodType]]
    plugins: Dict[str, ModelPlugin] = {}

    def __init__(
        self,
        config: Union[List[str], str] = "",
        mode: str = "train",
        weights: str = None,
        logger: logging.Logger = None,
        verbose: int = 2,
        **kwargs
    ):
        """Initializes the EdnaML object with the associated config, mode, weights, and verbosity.
        Sets up the logger, as well as local logger save directory.

        Args:
            config (str, optional): Path to the Edna config file. Defaults to "config.yaml".
            mode (str, optional): Either `train` or `test`. Defaults to "train".
            weights (str, optional): The path to the weights file. Defaults to None.
            logger (logging.Logger, optional): A logger. If no logger is provided, EdnaML will construct its own. Defaults to None.
            verbose (int, optional): Logging verbosity. Defaults to 2.

        Kwargs:
            load_epoch: If you want EdnaML to load a model from a specific saved epoch. The path construction will be inferred from the config file
            add_filehandler: If you want logger to write to log file
            add_streamhandler: If you want logger to write to stdout
            logger_save_name: If you want logger name to be something other than default constructed name from config
            test_only: Under this condition, if in test mode, only model and dataloaders will be created. Optimizers, schedulers will be empty.

        """

        if type(config) is str:
            self.config = config
        elif type(config) is list:
            self.config = []
            for cpath in config:
                if type(cpath) is str:
                    self.config.append(cpath)
                elif type(cpath) is list:
                    self.config += cpath
                else:
                    ValueError("config MUST be list or string")
        else:
            raise ValueError("config MUST be list or string")
        
        self.mode = mode
        self.weights = weights
        self.pretrained_weights = None
        self.verbose = verbose
        self.gpus = torch.cuda.device_count()
        # TODO Deal with extensions
        if type(self.config) is str:
            self.cfg = EdnaMLConfig([self.config], **kwargs)
        else:
            self.cfg = EdnaMLConfig(self.config, **kwargs)

        self.saveMetadata = SaveMetadata(
            self.cfg, **kwargs
        )  # <-- for changing the logger name...
        os.makedirs(self.saveMetadata.MODEL_SAVE_FOLDER, exist_ok=True)

        self.logger = self.buildLogger(logger=logger, **kwargs)
        self.previous_stop = -1
        self.load_epoch = kwargs.get("load_epoch", None)
        self.test_only = kwargs.get("test_only", False)
        if self.test_only and self.mode == "train":
            raise ValueError(
                "Cannot have `test_only` and be in training mode. Switch to"
                " `test` mode."
            )
        self.resetQueueArray:List[MethodType] = [self.resetCrawlerQueues, self.resetGeneratorQueues, self.resetModelQueues, self.resetOptimizerQueues, self.resetLossBuilderQueue, self.resetStorageQueues, self.resetTrainerQueue]
        self.resetQueueArray += self.addResetQueues()
        self.resetQueues()

        self.decorator_reference = {
            "crawler": self.addCrawlerClass,
            "model": self.addModelClass,
            "trainer": self.addTrainerClass,
            #"storage": self.addStorageClass,
            "generator": self.addGeneratorClass,
            "model_plugin": self.addPlugins,
        }

    def addResetQueues(self):
        return []
    def resetCrawlerQueues(self):
        self._crawlerClassQueue = None
        self._crawlerArgsQueue = None
        self._crawlerArgsQueueFlag = False
        self._crawlerClassQueueFlag = False
        self._crawlerInstanceQueue = None
        self._crawlerInstanceQueueFlag = False
    def resetGeneratorQueues(self):
        self._generatorClassQueue = None
        self._generatorArgsQueue = None
        self._generatorArgsQueueFlag = False
        self._generatorClassQueueFlag = False
        self._trainGeneratorInstanceQueue = None
        self._trainGeneratorInstanceQueueFlag = False
        self._testGeneratorInstanceQueue = None
        self._testGeneratorInstanceQueueFlag = False
    def resetModelQueues(self):
        self._modelBuilderQueue = None
        self._modelBuilderQueueFlag = False
        self._modelConfigQueue = None
        self._modelConfigQueueFlag = False
        self._modelQueue = None
        self._modelQueueFlag = False
        self._modelClassQueue = None
        self._modelClassQueueFlag = False
        self._modelArgsQueue = None
        self._modelArgsQueueFlag = False
    def resetOptimizerQueues(self):
        self._optimizerQueue = []
        self._optimizerNameQueue = []
        self._optimizerQueueFlag = False
    def resetLossBuilderQueue(self):
        self._lossBuilderQueue = []
        self._lossBuilderQueueFlag = False
    def resetTrainerQueue(self):
        self._trainerClassQueue = None
        self._trainerClassQueueFlag = False
        self._trainerInstanceQueue = None
        self._trainerInstanceQueueFlag = False
    def resetStorageQueues(self):
        self._storageInstanceQueueFlag = False
        self._storageClassQueueFlag = False
        self._storageInstanceQueue = None
        self._storageClassQueue = None


    def resetQueues(self):
        """Resets the `apply()` queue"""
        self.logger.debug("Resetting declarative queues.")
        for queue_function in self.resetQueueArray:
            queue_function()

    def recordVars(self):
        """recordVars() completes initial setup, allowing you to proceed with the core ml pipeline
        of dataloading, model building, etc
        """

        self.epochs = self.cfg.EXECUTION.EPOCHS
        self.skipeval = self.cfg.EXECUTION.SKIPEVAL
        self.step_verbose = self.cfg.LOGGING.STEP_VERBOSE
        self.save_frequency = self.cfg.SAVE.SAVE_FREQUENCY
        self.test_frequency = self.cfg.EXECUTION.TEST_FREQUENCY
        self.fp16 = self.cfg.EXECUTION.FP16

    def downloadModelWeights(self):
        """Downloads model weights specified in the configuration if `weights` were not passed into EdnaML and if model weights are supported.

        TODO -- do not throw error for no weights or missing base, if this is a new architecture to be trained from scratch
        Raises:
            Warning: If there are no pre-downloaded weights, and the model architecture is unsupported
        """
        if self.weights is not None:
            self.logger.info(
                "Not downloading weights. Weights path already provided."
            )

        if self.mode == "train":
            self._download_weights_from_base(self.cfg.MODEL.MODEL_BASE)
        else:
            if self.weights is None:
                warnings.warn(
                    "Mode is `test` but weights is `None`. This will cause"
                    " issues when EdnaML attempts to load weights"
                )

    def _download_weights_from_base(self, model_base: str):
        """Downloads weights from a model_base parameter directly from pytorch's servers.

        Args:
            model_base (str): A supported `model_base`, e.g. resnet18, resnet34. See `utils.web.model_weights`.
        """
        from ednaml.utils import model_weights

        if model_base in model_weights:
            if os.path.exists(model_weights[model_base][1]):
                pass
            else:
                self.logger.info(
                    "Model weights file {} does not exist. Downloading.".format(
                        model_weights[model_base][1]
                    )
                )
                ednaml.utils.web.download(
                    model_weights[model_base][1], model_weights[model_base][0]
                )
            self.pretrained_weights = model_weights[model_base][1]
        else:
            warnings.warn(
                "Model %s is not available. Please choose one of the following:"
                " %s if you want to load pretrained weights"
                % (model_base, str(model_weights.keys()))
            )

    def apply(self, **kwargs):
        """Applies the internal configuration for EdnaML"""
        self.printConfiguration()
        self.downloadModelWeights()
        self.setPreviousStop()
        self.buildStorage() # this is where -- same as other build things...  

        self.buildDataloaders()

        self.buildModel()
        self.loadWeights()
        if not kwargs.get("skip_model_summary", False):
            if self.cfg.LOGGING.INPUT_SIZE is not None:
                if kwargs.get("input_size", None) is None:
                    kwargs["input_size"] = self.cfg.LOGGING.INPUT_SIZE
            self.getModelSummary(**kwargs) 
        self.buildOptimizer()
        self.buildScheduler()

        self.buildLossArray()
        self.buildLossOptimizer()
        self.buildLossScheduler()

        self.buildStorage()
        self.buildTrainer()

        self.resetQueues()

    def addStorage(self, storage: BaseStorage):
        self.logger.debug("Added storage: %s"%storage.__class__.__name__)
        self._storageInstanceQueue = storage
        self._storageInstanceQueueFlag = True
    
    def addStorageClass(self, storage_class: Type[BaseStorage]):
        self.logger.debug("Added storage class: %s"%storage_class.__name__)
        self._storageClassQueue = storage_class
        self._storageClassQueueFlag = True

    def buildStorage(self):  
        # Options: a class was passed through the edna add functionality
        # Alternatively, we have specified a specific class in the storage.type section
        # Alternatively, we have specified a type, but that does bot exist as a built in

        # So, first we check if a class or instance was directly passed.
        # If nothing has been passed, then we try to locate a built-in version based on storage.type
        # If that doesn't work, then we can throw an error...

        if self._storageClassQueueFlag:
            storage = self._storageClassQueue
        else:
            storage: Type[BaseStorage] = locate_class(
                subpackage="storage", classpackage=self.cfg.STORAGE.TYPE
            )
            self.logger.info(
                "Loaded {} from {} to build Storage".format(
                    self.cfg.STORAGE.TYPE, "ednaml.storage"
                )
            )

        if self._storageInstanceQueueFlag:
            self.storage = self._storageInstanceQueue
        else:
            self.storage = storage(type=self.cfg.STORAGE.TYPE,
                                    url=self.cfg.STORAGE.URL,
                                    **self.cfg.STORAGE.STORAGE_ARGS)

    def train(self, **kwargs):
        self.trainer.train(continue_epoch=self.previous_stop + 1, **kwargs)  #

    def eval(self):
        return self.trainer.evaluate()

    def addTrainerClass(self, trainerClass: Type[BaseTrainer]):
        self.logger.debug("Added trainer class: %s"%trainerClass.__name__)
        self._trainerClassQueue = trainerClass
        self._trainerClassQueueFlag = True

    def addTrainer(self, trainer: BaseTrainer):
        self._trainerInstanceQueue = trainer
        self._trainerInstanceQueueFlag = True

    def buildTrainer(self):
        """Builds the EdnaML trainer and sets it up"""
        if self._trainerClassQueueFlag:
            ExecutionTrainer = self._trainerClassQueue
        else:
            ExecutionTrainer: Type[BaseTrainer] = locate_class(
                subpackage="trainer", classpackage=self.cfg.EXECUTION.TRAINER
            )
            self.logger.info(
                "Loaded {} from {} to build Trainer".format(
                    self.cfg.EXECUTION.TRAINER, "ednaml.trainer"
                )
            )

        if self._trainerInstanceQueueFlag:
            self.trainer = self._trainerInstanceQueue
        else:
            self.trainer = ExecutionTrainer(
                model=self.model,
                loss_fn=self.loss_function_array,
                optimizer=self.optimizer,
                loss_optimizer=self.loss_optimizer_array,
                scheduler=self.scheduler,
                loss_scheduler=self.loss_scheduler,
                train_loader=self.train_generator.dataloader,
                test_loader=self.test_generator.dataloader,
                epochs=self.cfg.EXECUTION.EPOCHS,
                skipeval=self.cfg.EXECUTION.SKIPEVAL,
                logger=self.logger,
                crawler=self.crawler,
                config=self.cfg,
                labels=self.labelMetadata,
                storage=self.storage,
                **self.cfg.EXECUTION.TRAINER_ARGS
            )
            self.trainer.apply(
                step_verbose=self.cfg.LOGGING.STEP_VERBOSE,
                save_frequency=self.cfg.SAVE.SAVE_FREQUENCY,
                test_frequency=self.cfg.EXECUTION.TEST_FREQUENCY,
                save_directory=self.saveMetadata.MODEL_SAVE_FOLDER,
                save_backup=self.cfg.SAVE.DRIVE_BACKUP,
                backup_directory=self.saveMetadata.CHECKPOINT_DIRECTORY,
                gpus=self.gpus,
                fp16=self.cfg.EXECUTION.FP16,
                model_save_name=self.saveMetadata.MODEL_SAVE_NAME,
                logger_file=self.saveMetadata.LOGGER_SAVE_NAME,
            )

    def setPreviousStop(self):
        """Sets the previous stop"""
        self.previous_stop = self.getPreviousStop()

    def getPreviousStop(self) -> int:
        """Gets the previous stop, if any, of the trainable model by checking local save directory, as well as a network directory."""
        if self.cfg.SAVE.DRIVE_BACKUP:
            fl_list = glob.glob(
                os.path.join(self.saveMetadata.CHECKPOINT_DIRECTORY, "*.pth")
            )
        else:
            fl_list = glob.glob(
                os.path.join(self.saveMetadata.MODEL_SAVE_FOLDER, "*.pth")
            )
        _re = re.compile(r".*epoch([0-9]+)\.pth")
        previous_stop = [
            int(item[1])
            for item in [_re.search(item) for item in fl_list]
            if item is not None
        ]
        if len(previous_stop) == 0:
            self.logger.info(
                "No previous stop detected. Will start from epoch 0"
            )
            return -1
        else:
            self.logger.info(
                "Previous stop detected. Will attempt to resume from epoch %i"
                % max(previous_stop)
            )
            return max(previous_stop)

    def addOptimizer(
        self, optimizer: torch.optim.Optimizer, parameter_group="opt-1"
    ):
        self._optimizerQueue.append(optimizer)
        self._optimizerNameQueue.append(parameter_group)
        self._optimizerQueueFlag = True

    def buildOptimizer(self):
        """Builds the optimizer for the model"""
        if self.test_only:
            self.optimizer = []
            self.logger.info(
                "Skipping optimizer building step in `test_only` mode"
            )
        else:
            if self._optimizerQueueFlag:
                self.optimizer = [item for item in self._optimizerQueue]
            else:
                optimizer_builder: Type[BaseOptimizer] = locate_class(
                    subpackage="optimizer",
                    classpackage=self.cfg.EXECUTION.OPTIMIZER_BUILDER,
                )
                self.logger.info(
                    "Loaded {} from {} to build Optimizer model".format(
                        self.cfg.EXECUTION.OPTIMIZER_BUILDER, "ednaml.optimizer"
                    )
                )

                # Optimizers are in a list...
                OPT_array = [
                    optimizer_builder(
                        name=optimizer_item.OPTIMIZER_NAME,
                        optimizer=optimizer_item.OPTIMIZER,
                        base_lr=optimizer_item.BASE_LR,
                        lr_bias=optimizer_item.LR_BIAS_FACTOR,
                        weight_decay=optimizer_item.WEIGHT_DECAY,
                        weight_bias=optimizer_item.WEIGHT_BIAS_FACTOR,
                        opt_kwargs=optimizer_item.OPTIMIZER_KWARGS,
                        gpus=self.gpus,
                    )
                    for optimizer_item in self.cfg.OPTIMIZER
                ]
                self.optimizer = [
                    OPT.build(
                        self.model.getParameterGroup(
                            self.cfg.OPTIMIZER[idx].OPTIMIZER_NAME
                        )
                    )
                    for idx, OPT in enumerate(OPT_array)
                ]  # TODO make this a dictionary???
            self.logger.info("Built optimizer")

    def buildScheduler(self):
        """Builds the scheduler for the model"""
        if self.test_only:
            self.scheduler = []
            self.logger.info(
                "Skipping scheduler building step in `test_only` mode"
            )
        else:
            self.scheduler = [None] * len(self.cfg.SCHEDULER)
            for idx, scheduler_item in enumerate(self.cfg.SCHEDULER):
                try:  # We first check if scheduler is part of torch's provided schedulers.
                    scheduler = locate_class(
                        package="torch.optim",
                        subpackage="lr_scheduler",
                        classpackage=scheduler_item.LR_SCHEDULER,
                    )
                except (
                    ModuleNotFoundError,
                    AttributeError,
                ):  # If it fails, then we try to import from schedulers implemented in scheduler/ folder
                    scheduler = locate_class(
                        subpackage="scheduler",
                        classpackage=scheduler_item.LR_SCHEDULER,
                    )
                self.scheduler[idx] = scheduler(
                    self.optimizer[idx],
                    last_epoch=-1,
                    **scheduler_item.LR_KWARGS
                )
            self.logger.info("Built scheduler")



    def addLossBuilder(
        self, loss_list, loss_lambdas, loss_kwargs, loss_name, loss_label
    ):
        self._lossBuilderQueue.append(
            LossConfig(
                {
                    "LOSSES": loss_list,
                    "LAMBDAS": loss_lambdas,
                    "KWARGS": loss_kwargs,
                    "LABEL": loss_label,
                    "NAME": loss_name,
                }
            )
        )
        self._lossBuilderQueueFlag = True

    def buildLossArray(self):
        """Builds the loss function array using the LOSS list in the provided configuration"""
        if self.test_only:
            self.loss_function_array = []
            self.logger.info(
                "Skipping loss function array building step in `test_only` mode"
            )
        else:
            self.loss_function_array = [
                ClassificationLossBuilder(
                    loss_functions=loss_item.LOSSES,
                    loss_lambda=loss_item.LAMBDAS,
                    loss_kwargs=loss_item.KWARGS,
                    name=loss_item.NAME,  # get("NAME", None),
                    label=loss_item.LABEL,  # get("LABEL", None),
                    metadata=self.labelMetadata,
                    **{"logger": self.logger}
                )
                for loss_item in self.cfg.LOSS
            ]
            if self._lossBuilderQueueFlag:
                self.loss_function_array += [
                    ClassificationLossBuilder(
                        loss_functions=loss_item.LOSSES,
                        loss_lambda=loss_item.LAMBDAS,
                        loss_kwargs=loss_item.KWARGS,
                        name=loss_item.NAME,  # get("NAME", None),
                        label=loss_item.LABEL,  # get("LABEL", None),
                        metadata=self.labelMetadata,
                        **{"logger": self.logger}
                    )
                    for loss_item in self._lossBuilderQueue
                ]
            self.logger.info("Built loss function")

    def buildLossOptimizer(self):
        """Builds the Optimizer for loss functions, if the loss functions have learnable parameters (e.g. proxyNCA loss)

        self.loss_function_array contains a list of LossBuilders. Each LossBuilder
        is for a specific output. Here, we build an array of StandardLossOptimizers,
        one StandardLossOptimizer for each LossBuilder. Each StandardLossOptimizer
        takes as input the same arguments as an Optimizer. However, the name parameter
        should be the name of the LossBuilder it is targeting.

        If there is no LOSS_OPTIMIZER section in the configuration, the EdnaML config creates a default
        LOSS_OPTIMIZER, whose parameters we will use.

        """
        if self.test_only:
            self.loss_optimizer_array = []
            self.logger.info(
                "Skipping loss-scheduler building step in `test_only` mode"
            )
        else:
            loss_optimizer_name_dict = {
                loss_optim_item.OPTIMIZER_NAME: loss_optim_item
                for loss_optim_item in self.cfg.LOSS_OPTIMIZER
            }
            initial_key = list(loss_optimizer_name_dict.keys())[0]
            LOSS_OPT: List[StandardLossOptimizer] = [None] * len(
                self.loss_function_array
            )
            for idx, loss_builder in enumerate(self.loss_function_array):
                if loss_builder.loss_labelname in loss_optimizer_name_dict:
                    # Means we have an optimizer corresponding to this loss
                    lookup_key = loss_builder.loss_labelname
                else:  # We will use the first optimizer (either default or otherwise, etc, for this)
                    lookup_key = initial_key
                LOSS_OPT[idx] = StandardLossOptimizer(
                    name=loss_optimizer_name_dict[lookup_key].OPTIMIZER_NAME,
                    optimizer=loss_optimizer_name_dict[lookup_key].OPTIMIZER,
                    base_lr=loss_optimizer_name_dict[lookup_key].BASE_LR,
                    lr_bias=loss_optimizer_name_dict[lookup_key].LR_BIAS_FACTOR,
                    gpus=self.gpus,
                    weight_bias=loss_optimizer_name_dict[
                        lookup_key
                    ].WEIGHT_BIAS_FACTOR,
                    weight_decay=loss_optimizer_name_dict[
                        lookup_key
                    ].WEIGHT_DECAY,
                    opt_kwargs=loss_optimizer_name_dict[
                        lookup_key
                    ].OPTIMIZER_KWARGS,
                )

            # Note: build returns None if there are no differentiable parameters
            self.loss_optimizer_array = [
                loss_opt.build(loss_builder=self.loss_function_array[idx])
                for idx, loss_opt in enumerate(LOSS_OPT)
            ]
            self.logger.info("Built loss optimizer")

    def buildLossScheduler(self):
        """Builds the scheduler for the loss functions, if the functions have learnable parameters and corresponding optimizer."""
        if self.test_only:
            self.loss_scheduler = []
            self.logger.info(
                "Skipping loss-scheduler building step in `test_only` mode"
            )
        else:
            loss_scheduler_name_dict = {
                loss_schedule_item.SCHEDULER_NAME: loss_schedule_item
                for loss_schedule_item in self.cfg.LOSS_SCHEDULER
            }
            initial_key = list(loss_scheduler_name_dict.keys())[0]
            self.loss_scheduler = [None] * len(self.loss_optimizer_array)

            for idx, loss_optimizer in enumerate(self.loss_optimizer_array):
                if (
                    loss_optimizer is not None
                ):  # In case loss has differentiable parameters, so the optimizer is not None...we look for the loss name
                    if (
                        self.loss_function_array[idx].loss_labelname
                        in loss_scheduler_name_dict
                    ):
                        lookup_key = self.loss_function_array[
                            idx
                        ].loss_labelname
                    else:
                        lookup_key = initial_key

                    try:  # We first check if scheduler is part of torch's provided schedulers.
                        loss_scheduler = importlib.import_module(
                            loss_scheduler_name_dict[lookup_key].LR_SCHEDULER,
                            package="torch.optim.lr_scheduler",
                        )
                    except (
                        ModuleNotFoundError,
                        AttributeError,
                    ):  # If it fails, then we try to import from schedulers implemented in scheduler/ folder
                        loss_scheduler = importlib.import_module(
                            loss_scheduler_name_dict[lookup_key].LR_SCHEDULER,
                            package="ednaml.scheduler",
                        )
                    self.loss_scheduler[idx] = loss_scheduler(
                        loss_optimizer,
                        last_epoch=-1,
                        **loss_scheduler_name_dict[lookup_key].LR_KWARGS
                    )
                self.logger.info("Built loss scheduler")

    def _setModelTestMode(self):
        """Sets model to test mode if EdnaML is in testing mode"""
        if self.mode == "test":
            self.model.eval()

    def _setModelTrainMode(self):
        """Sets the model to train mode if EdnaML is in training mode"""
        if self.mode == "train":
            self.model.train()

    def _covert_model_kwargs(self) -> Dict[str, int]:
        """Converts the model_kwargs inside config into the correct format, depending on whether it is provided directly in yaml format, or as a json string

        Returns:
            Dict[str,Union[str,int]]: Corrected model_kwargs dictionary
        """

        if (
            type(self.cfg.MODEL.MODEL_KWARGS) is dict
        ):  # Compatibility with old configs. TODO fix all old configs.
            model_kwargs_dict = self.cfg.MODEL.MODEL_KWARGS
        elif type(self.cfg.MODEL.MODEL_KWARGS) is None:
            model_kwargs_dict = {}
        elif type(self.cfg.MODEL.MODEL_KWARGS) is str:
            raise ValueError("Outdated model_kwargs as str")
            # model_kwargs_dict = json.loads(self.cfg.MODEL.MODEL_KWARGS)
        return model_kwargs_dict

    def buildModel(self):
        """Builds an EdnaML model using the configuration. If there are pretrained weights, they are provided through the config to initialize the model."""
        if self._modelBuilderQueueFlag:
            model_builder = self._modelBuilderQueue
        else:
            model_builder = locate_class(
                subpackage="models", classpackage=self.cfg.MODEL.BUILDER
            )
        self.logger.info(
            "Loaded {} from {} to build model".format(
                self.cfg.MODEL.BUILDER, "ednaml.models"
            )
        )

        if self._modelConfigQueueFlag:
            self.cfg.MODEL = self._modelConfigQueue

        # model_kwargs = self._covert_model_kwargs()

        # TODO!!!!!!!
        if self._modelQueueFlag:
            self.model = self._modelQueue
        else:
            if self._modelArgsQueueFlag:
                self.cfg.MODEL.MODEL_KWARGS = self._modelArgsQueue

            if self._modelClassQueueFlag:
                self.cfg.MODEL.MODEL_ARCH = self._modelClassQueue.__name__
                arch = self._modelClassQueue
            else:
                arch = self.cfg.MODEL.MODEL_ARCH
            self.model: ModelAbstract = model_builder(
                arch=arch,
                base=self.cfg.MODEL.MODEL_BASE,
                weights=self.pretrained_weights,
                metadata=self.labelMetadata,
                normalization=self.cfg.MODEL.MODEL_NORMALIZATION,
                parameter_groups=self.cfg.MODEL.PARAMETER_GROUPS,
                **self.cfg.MODEL.MODEL_KWARGS
            )
        self.logger.info(
            "Finished instantiating model with {} architecture".format(
                self.cfg.MODEL.MODEL_ARCH
            )
        )
        self.model._logger = self.logger
        self.logger.info("Adding plugins after constructing model")
        self.addPluginsToModel()

    # Add model hooks here, since it is a ModelAbstract
    def addPlugins(self, plugin_class_list: List[Type[ModelPlugin]], **kwargs):
        """add plugins to the plugins class queue...
        ideally this is called before add model
        then add model can add the plugins in the queue to the model...

        NOTE -- any plugins are added through the config file, specifically, with the plugin arguments. 
        This is because multiple plugins can use the same class, but variations of it.
        For example, one might want to train KMP under l2 and cos, and use then simultaneously...

        So, plugins class queue exists only as a lookup for plugin classes...if a correspondiing enttry is NOT in MODEL_PLUGINS in the config,
        it will never be added into Edna/Deploy...
        """
        for plugin in plugin_class_list:
            self.logger.debug("Added custom plugin: %s"%plugin.__name__)
            self.plugins[plugin.__name__] = plugin  # order matters; can replace...though shouldn't matter too much...
        
        # The follow will not happen, logically, i think, because model is ONLY defined in self.buildModel(), and by that point, one is already calling apply()
        # Adding plugins after the fact...????? Don't think so...
        # And in any case, we should make apply() truely declarative, so that one can simply apply() again after adding plugins
        #if self.model is not None:
        #    # model is defined. We will add plugins to queue, and then call the model's add plugin bit...
        #    self.logger.info("Model already constructed. Adding plugins.")
        #    self.addPluginsToModel()

    def addPluginsToModel(self):
        # Here, we iterate through plugins IN the config file, then import their classes, and add them to the model...
        for plugin_save_name in self.cfg.MODEL_PLUGIN:
            # Now, we need to locate the plugin class...check first if it is in plugins...
            if self.cfg.MODEL_PLUGIN[plugin_save_name].PLUGIN not in self.plugins:
                # now we need to locate the class in builtins
                plugin = locate_class(
                    subpackage="plugins",
                    classpackage=self.cfg.MODEL_PLUGIN[plugin_save_name].PLUGIN,
                    classfile=self.cfg.MODEL_PLUGIN[plugin_save_name].PLUGIN
                )
            else:
                plugin = self.plugins[self.cfg.MODEL_PLUGIN[plugin_save_name].PLUGIN]
            
            self.model.addPlugin(
                plugin = plugin,
                plugin_name = plugin_save_name,
                plugin_kwargs = self.cfg.MODEL_PLUGIN[plugin_save_name].PLUGIN_KWARGS
            )

        #for plugin in self.plugins:
        #    self.model.addPlugin(plugin, 
        #                    plugin_name = self.cfg.MODEL_PLUGIN[plugin.name].PLUGIN_NAME, 
        #                    plugin_kwargs = self.cfg.MODEL_PLUGIN[plugin.name].PLUGIN_KWARGS)
            
    def addModelBuilder(
        self, model_builder: Type[Callable], model_config: ModelConfig = None
    ):
        self._modelBuilderQueue = model_builder
        self._modelBuilderQueueFlag = True
        if model_config is not None:
            self._modelConfigQueue = model_config
            self._modelConfigQueueFlag = True

    def addModel(self, model: ModelAbstract):
        self._modelQueue = model
        self._modelQueueFlag = True

    def addModelClass(self, model_class: Type[ModelAbstract], **kwargs):
        self.logger.debug("Added model class: %s"%model_class.__name__)
        self._modelClassQueue = model_class
        self._modelClassQueueFlag = True
        self._modelArgsQueue = kwargs
        if len(self._modelArgsQueue) > 0:
            self._modelArgsQueueFlag = True

    def loadWeights(self):
        """If in `test` mode, load weights from weights path. Otherwise, partially load what is possible from given weights path, if given.
        Note that for training mode, weights are downloaded from pytoch to be loaded if pretrained weights are desired.
        """
        if self.mode == "test":
            if self.weights is None:
                self.logger.info(
                    "No saved model weights provided. Inferring weights path."
                )
                if self.load_epoch is not None and type(self.load_epoch) is int:
                    self.weights = self.getModelWeightsFromEpoch(
                        self.load_epoch
                    )
                    self.logger.info(
                        "Using weights from provided epoch %i, at path %s."
                        % (self.load_epoch, self.weights)
                    )
                else:
                    if self.previous_stop < 0:
                        self.logger.info(
                            "No previous stop exists. Not loading weights."
                        )
                    else:
                        self.weights = self.getModelWeightsFromEpoch(
                            self.previous_stop
                        )
                        self.logger.info(
                            "Using weights from last saved epoch %i, at"
                            " path %s." % (self.previous_stop, self.weights)
                        )
            if self.weights is not None:    # we have this, because previous if-block might update weights path
                self.model.load_state_dict(torch.load(self.weights))
        else:
            if self.weights is None:
                self.logger.info("No saved model weights provided.")
            else:
                if (
                    self.weights != ""
                ):  # Load weights if train and starting from a another model base...
                    self.logger.info(
                        "Commencing partial model load from {}".format(
                            self.weights
                        )
                    )
                    self.model.partial_load(self.weights)
                    self.logger.info(
                        "Completed partial model load from {}".format(
                            self.weights
                        )
                    )
                else:
                    self.logger.info(
                        "In train mode, but no weights provided to load into"
                        " model. This means either model will be initialized"
                        " randomly, or weights loading occurs inside the model."
                    )

    def getModelSummary(self, input_size=None, dtypes=None, **kwargs):
        """Gets the model summary using `torchinfo` and saves it as a ModelStatistics object"""
        # add bookkeeping
        # does config has input size? if it does use that, but prepend batch size .
        # if it doesn't have it, use input size in arguments
        # TODO possibly move this into trainer...?
        # or at least deal with potential mlti-gpu scenario...
        device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        self.model.to(device)
        # change below statement according to line 722
        # default for input size is None/null
        try:
            if input_size is None:
                input_size = (
                    self.cfg.TRAIN_TRANSFORMATION.BATCH_SIZE,
                    self.cfg.TRAIN_TRANSFORMATION.INPUT_SIZE,
                ) # INPUT SIZE SHOULD HAVE A VALUE
            print("INPUT SIZE ==== ",input_size)

            self.model_summary = summary(
                self.model,
                input_size=input_size,
                col_names=[
                    "input_size",
                    "output_size",
                    "num_params",
                    "kernel_size",
                    "mult_adds",
                ],
                depth=3,
                mode="train",
                verbose=0,
                dtypes=dtypes,
            )
            self.logger.info(str(self.model_summary))
        except KeyboardInterrupt:   # TODO check which exception is actually raised in summary and maybe have a better way to deal with this...
            raise
        except Exception as e:
            import traceback
            self.logger.info("Model Summary retured the following error:")
            self.logger.info(traceback.format_exc())


    #------------------------------GENERAL ADD--------------------------------------------------------------------
    def _add(self, file_or_module_path):
        # We are given a file, which contains several class loaded through decorators when the file is imported.
        imported_file = path_import(file_or_module_path)
        # Once we have imported, the files are registered in ednaml.core.decorators.REGISTERED_EDNA_COMPONENTS
        lookup_path = os.path.abspath(file_or_module_path)
        for keyvalue in ednaml.core.decorators.REGISTERED_EDNA_COMPONENTS[lookup_path]:
            if keyvalue in self.decorator_reference:
                if type(ednaml.core.decorators.REGISTERED_EDNA_COMPONENTS[lookup_path][keyvalue]) is list:
                  for i in range(len(ednaml.core.decorators.REGISTERED_EDNA_COMPONENTS[lookup_path][keyvalue])):
                    self.logger.info("Adding a {ftype}, from {src}, with inferred name {inf}".format(
                    ftype=keyvalue,
                    src = lookup_path,
                    inf = ednaml.core.decorators.REGISTERED_EDNA_COMPONENTS[lookup_path][keyvalue][i].__name__
                ))
                else:
                  self.logger.info("Adding a {ftype}, from {src}, with inferred name {inf}".format(
                      ftype=keyvalue,
                      src = lookup_path,
                      inf = ednaml.core.decorators.REGISTERED_EDNA_COMPONENTS[lookup_path][keyvalue].__name__
                  ))
                self.decorator_reference[keyvalue](
                    ednaml.core.decorators.REGISTERED_EDNA_COMPONENTS[lookup_path][keyvalue]
                )
            else:
                warnings.warn(
                    "keyvalue %s in REGISTERED_EDNA_COMPONENTS %s is not available in self.decorator_reference. Not adding."%(keyvalue, 
                    str(ednaml.core.decorators.REGISTERED_EDNA_COMPONENTS[lookup_path][keyvalue]))
                )
    def add(self, file_or_module_path):
        if type(file_or_module_path) is list:
            for file_or_module in file_or_module_path:
                self._add(file_or_module)
        else:
            self._add(file_or_module_path)
        
    # ----------------------------------------------   DATAREADERS   ----------------------------------------------
    def addCrawlerClass(self, crawler_class: Type[Crawler], **kwargs):
        """Adds a crawler class to the EdnaML `apply()` queue. This will be applied to the configuration when calling `apply()`

        The crawler class is added to the internal datareader instance through `apply()`. Then,
        the buildTrainDataloader() and buildTestDataloader() can take instances of this class
        to crawl the dataset and yield batches.

        Args:
            crawler_class (Type[Crawler]): _description_
        """
        self._crawlerClassQueue = crawler_class
        self._crawlerArgsQueue = kwargs
        if len(self._crawlerArgsQueue) > 0:
            self._crawlerArgsQueueFlag = True
        self._crawlerClassQueueFlag = True

    def addCrawler(self, crawler_instance: Crawler):
        """Adds a crawler instance to the EdnaML `apply()` queue.

        Args:
            crawler_instance (Crawler): _description_
        """
        self._crawlerInstanceQueue = crawler_instance
        self._crawlerInstanceQueueFlag = True

    def addGeneratorClass(self, generator_class: Type[Generator], **kwargs):
        """Adds a generator class to the EdnaML `apply()` queue. This will be applied to the configuration when calling `apply()`

        The generator class is added to the internal datareader instance through `apply()`. Then,
        the buildTrainDataloader() and buildTestDataloader() can take instances of this class
        to crawl the dataset and yield batches.

        Args:
            generator_class (Type[Generator]): _description_
        """
        self._generatorClassQueue = generator_class
        self._generatorArgsQueue = kwargs
        if len(self._generatorArgsQueue) > 0:
            self._generatorArgsQueueFlag = True
        self._generatorClassQueueFlag = True

    def addTrainGenerator(self, generator: Generator):
        """Adds a generator instance to the EdnaML `apply()` queue.

        Args:
            generator_class (Type[Generator]): _description_
        """
        self._trainGeneratorInstanceQueue = generator
        self._trainGeneratorInstanceQueueFlag = True

    def addTestGenerator(self, generator: Generator):
        """Adds a generator instance to the EdnaML `apply()` queue.

        Args:
            generator_class (Type[Generator]): _description_
        """
        self._testGeneratorInstanceQueue = generator
        self._testGeneratorInstanceQueueFlag = True

    def buildDataloaders(self):
        """Sets up the datareader classes and builds the train and test dataloaders"""

        data_reader: Type[DataReader] = locate_class(
            package="ednaml",
            subpackage="datareaders",
            classpackage=self.cfg.EXECUTION.DATAREADER.DATAREADER,
        )
        data_reader_instance = data_reader()
        self.logger.info("Reading data with DataReader %s" % data_reader_instance.name)
        self.logger.info("Default CRAWLER is %s"%data_reader_instance.CRAWLER)
        self.logger.info("Default DATASET is %s"%data_reader_instance.DATASET)
        self.logger.info("Default GENERATOR is %s"%data_reader_instance.GENERATOR)
        # Update the generator...if needed
        if self._generatorClassQueueFlag:
            self.logger.info("Updating GENERATOR to queued class %s"%self._generatorClassQueue.__name__)
            data_reader_instance.GENERATOR = self._generatorClassQueue
            if self._generatorArgsQueueFlag:
                self.cfg.EXECUTION.DATAREADER.GENERATOR_ARGS = (
                    self._generatorArgsQueue
                )
        else:
            if (
                self.cfg.EXECUTION.DATAREADER.GENERATOR is not None
            ):
                self.logger.info("Updating GENERATOR using config specification to %s"%self.cfg.EXECUTION.DATAREADER.GENERATOR)
                data_reader_instance.GENERATOR = locate_class(
                    package="ednaml",
                    subpackage="generators",
                    classpackage=self.cfg.EXECUTION.DATAREADER.GENERATOR,
                )

        if self._crawlerClassQueueFlag: #here it checkes whether class flag is set, if it is then replace the build in class with custom class
            self.logger.info("Updating CRAWLER to %s"%self._crawlerClassQueue.__name__)
            data_reader_instance.CRAWLER = self._crawlerClassQueue
            if self._crawlerArgsQueueFlag: #check args also
                self.cfg.EXECUTION.DATAREADER.CRAWLER_ARGS = (
                    self._crawlerArgsQueue
                )

        if self._crawlerInstanceQueueFlag:
            self.crawler = self._crawlerInstanceQueue
        else:
            self.crawler = self._buildCrawlerInstance(
                data_reader=data_reader_instance
            )

        self.buildTrainDataloader(data_reader_instance, self.crawler)
        self.buildTestDataloader(data_reader_instance, self.crawler)

    def _buildCrawlerInstance(self, data_reader: DataReader) -> Crawler:
        """Builds a Crawler instance from the data_reader's provided crawler class in `data_reader.CRAWLER`

        Args:
            data_reader (DataReader): A DataReader class

        Returns:
            Crawler: A Crawler instanece for this experiment
        """
        return data_reader.CRAWLER(
            logger=self.logger, **self.cfg.EXECUTION.DATAREADER.CRAWLER_ARGS
        )

    def buildTrainDataloader(
        self, data_reader: DataReader, crawler_instance: Crawler
    ):
        """Builds a train dataloader instance given the data_reader class and a crawler instance that has been initialized

        Args:
            data_reader (DataReader): A datareader class
            crawler_instance (Crawler): A crawler instance
        """
        if self._trainGeneratorInstanceQueueFlag:
            self.train_generator: Generator = self._trainGeneratorInstanceQueue
        else:
            #print(self.cfg.TRAIN_TRANSFORMATION)
            if self.mode != "test":
                self.train_generator: Generator = data_reader.GENERATOR( ##imp -- initialize generator
                    logger=self.logger,
                    gpus=self.gpus,
                    transforms=self.cfg.TRAIN_TRANSFORMATION,# train_transforms.args ## not imp.. all arguments are in args -- args is attribute which is storoing a dictionary
                    mode="train",
                    **self.cfg.EXECUTION.DATAREADER.GENERATOR_ARGS
                )

                self.train_generator.build( ## imp -- calls build method inside generator class
                    crawler_instance, 
                    batch_size=self.cfg.TRAIN_TRANSFORMATION.BATCH_SIZE, 
                    workers=self.cfg.TRAIN_TRANSFORMATION.WORKERS,
                    **self.cfg.EXECUTION.DATAREADER.DATASET_ARGS
                )
            else:
                self.logger.info(
                "Not creating `train_generator` in `test_only` mode."
                )
                self.train_generator = Generator(logger=self.logger)
        if self.mode != "test":
            self.logger.info(
                "Generated training data generator with %i training data points"
                % len(self.train_generator.dataset)
            )
            self.labelMetadata = self.train_generator.num_entities
            self.logger.info(
                "Running classification model with classes: %s"
                % str(self.labelMetadata.metadata)
            )
        else:
            self.logger.info(
                "Skipped generating training data, because EdnaML is in test"
                " mode."
            )

    def buildTestDataloader(
        self, data_reader: DataReader, crawler_instance: Crawler
    ):
        """Builds a test dataloader instance given the data_reader class and a crawler instance that has been initialized

        Args:
            data_reader (DataReader): A datareader class
            crawler_instance (Crawler): A crawler instance
        """
        if self._testGeneratorInstanceQueueFlag:
            self.test_generator: Generator = self._testGeneratorInstanceQueue
        else:
            self.test_generator: Generator = data_reader.GENERATOR(
                logger=self.logger,
                gpus=self.gpus,
                transforms=self.cfg.TEST_TRANSFORMATION,
                mode="test",
                **self.cfg.EXECUTION.DATAREADER.GENERATOR_ARGS
            )
            self.test_generator.build( 
                crawler_instance,
                batch_size=self.cfg.TEST_TRANSFORMATION.BATCH_SIZE,
                workers=self.cfg.TEST_TRANSFORMATION.WORKERS,
                **self.cfg.EXECUTION.DATAREADER.DATASET_ARGS
            )

        if self.mode == "test":
            self.labelMetadata = self.test_generator.num_entities
        self.logger.info("Generated test data/query generator")

    def buildLogger(
        self,
        logger: logging.Logger = None,
        add_filehandler: bool = True,
        add_streamhandler: bool = True,
        **kwargs
    ) -> logging.Logger:
        """Builds a new logger or adds the correct file and stream handlers to
        existing logger if it does not already have them.

        Args:
            logger (logging.Logger, optional): A logger.. Defaults to None.
            add_filehandler (bool, optional): Whether to add a file handler to the logger. If False, no file is created or appended to. Defaults to True.
            add_streamhandler (bool, optional): Whether to add a stream handler to the logger. If False, logger will not stream to stdout. Defaults to True.

        Returns:
            logging.Logger: A logger with file and stream handlers.
        """
        loggerGiven = True
        if logger is None:
            logger = logging.Logger(self.saveMetadata.MODEL_SAVE_FOLDER)
            loggerGiven = False

        logger_save_path = os.path.join(
            self.saveMetadata.MODEL_SAVE_FOLDER,
            self.saveMetadata.LOGGER_SAVE_NAME,
        )
        # Check for backup logger
        if self.cfg.SAVE.LOG_BACKUP:
            backup_logger = os.path.join(
                self.saveMetadata.CHECKPOINT_DIRECTORY,
                self.saveMetadata.LOGGER_SAVE_NAME,
            )
            if os.path.exists(backup_logger) and add_filehandler:
                print(
                    "Existing log file exists at network backup %s. Will"
                    " attempt to copy to local directory %s."
                    % (backup_logger, self.saveMetadata.MODEL_SAVE_FOLDER)
                )
                shutil.copy2(backup_logger, self.saveMetadata.MODEL_SAVE_FOLDER)
        if os.path.exists(logger_save_path) and add_filehandler:
            print(
                "Log file exists at %s. Will attempt to append there."
                % logger_save_path
            )

        streamhandler = False
        filehandler = False

        if logger.hasHandlers():
            for handler in logger.handlers():
                if isinstance(handler, logging.StreamHandler):
                    streamhandler = True
                if isinstance(handler, logging.FileHandler):
                    if handler.baseFilename == os.path.abspath(
                        logger_save_path
                    ):
                        filehandler = True

        if not loggerGiven:
            logger.setLevel(logging.DEBUG)

        if not filehandler and add_filehandler:
            fh = logging.FileHandler(
                logger_save_path, mode="a", encoding="utf-8"
            )
            fh.setLevel(self.logLevels[self.verbose])
            formatter = logging.Formatter(
                "%(asctime)s %(message)s", datefmt="%H:%M:%S"
            )
            fh.setFormatter(formatter)
            logger.addHandler(fh)

        if not streamhandler and add_streamhandler:
            cs = logging.StreamHandler()
            cs.setLevel(self.logLevels[self.verbose])
            cs.setFormatter(
                logging.Formatter("%(asctime)s %(message)s", datefmt="%H:%M:%S")
            )
            logger.addHandler(cs)

        return logger

    def log(self, message: str, verbose: int = 3):
        """Logs a message. TODO needs to be fixed.

        Args:
            message (str): Message to log
            verbose (int, optional): Logging verbosity. Defaults to 3.
        """
        self.logger.log(self.logLevels[verbose], message)

    def printConfiguration(self):
        """Prints the EdnaML configuration"""
        self.logger.info("*" * 40)
        self.logger.info("")
        self.logger.info("")
        self.logger.info("Using the following configuration:")
        self.logger.info(self.cfg.export())
        self.logger.info("")
        self.logger.info("")
        self.logger.info("*" * 40)

    def getModelWeightsFromEpoch(self, epoch=0):
        model_load = (
            self.saveMetadata.MODEL_SAVE_NAME + "_epoch%i" % epoch + ".pth"
        )
        if self.cfg.SAVE.DRIVE_BACKUP:
            self.logger.info("Loading model from drive backup.")
            model_load_path = os.path.join(
                self.saveMetadata.CHECKPOINT_DIRECTORY, model_load
            )
        else:
            self.logger.info("Loading model from local backup.")
            model_load_path = os.path.join(
                self.saveMetadata.MODEL_SAVE_FOLDER, model_load
            )
        if os.path.exists(model_load_path):
            return model_load_path
        return None

    def setModelWeightsFromEpoch(self, epoch=0):
        """Sets the internal model weights path using the provided epoch value.
        If there is no corresponding weights path to this epoch value, then
        sets the weights path to `None`

        Args:
            epoch (int, optional): The weights saved at this epoch. Defaults to 0.
        """
        self.weights = self.getModelWeightsFromEpoch(epoch=epoch)

    def loadEpoch(self, epoch=0):
        """Loads weights saved at a specific epoch into the current stored model
        in `self.model`. If no weights are saved for that epoch, logs this and
        does nothing.

        For the provided epoch, `loadEpoch` will check local save directory as
        well as backup save directory for file matching model name constructor
        from `self.saveMetadata`

        Args:
            epoch (int, optional): The epoch to load. If None, then will not do anything.
                If weights corresponding to this epoch do not exist, will not do anything. Defaults to 0.
        """
        model_load_path = self.getModelWeightsFromEpoch(epoch=epoch)
        if model_load_path is not None:
            self.model.load_state_dict(torch.load(model_load_path))
            self.logger.info(
                "Finished loading model state_dict from %s" % model_load_path
            )
        else:
            self.logger.info("No saved weights provided.")

    def deleteLocal(self):
        import shutil
        shutil.rmtree(self.saveMetadata.MODEL_SAVE_FOLDER)

    def setBackupToFalse(self):
        self.cfg.SAVE.DRIVE_BACKUP = False


from glob import glob
import random
import torch, os, shutil
from torch.utils.data import TensorDataset
from tqdm import tqdm
class AlbertDataset(torch.utils.data.Dataset):
    def __init__(self, logger, dataset, mode, transform=None, **kwargs):
        self.dataset = dataset  # list of tuples (text, labels, labels)
        self.logger = logger
        self.data_shuffle = kwargs.get("data_shuffle", True)
        if self.data_shuffle:
            random.shuffle(self.dataset)

        # Options
        self.cache = kwargs.get("cache", False)
        self.memcache = kwargs.get("memcache", False)

        self.shardcache = kwargs.get("shardcache", False)
        self.shardsize = kwargs.get("shardsize", 0)
        self.shardsaveindex = 0
        self.shard_replace = kwargs.get("shard_replace", False)
        self.shardpath = kwargs.get("shardpath", "datashard-artifacts")
        self.shardpath = mode+"-"+self.shardpath
        self.shardname = kwargs.get("shardname", "albert-shard") + "-"  #the dash
        self.base_shardpath = os.path.join(self.shardpath, self.shardname)
        self.shards_exist = False
        if os.path.exists(self.base_shardpath + "0.pt"):
            if self.shard_replace:
                self.logger.debug("Deleting existing shards")
                shutil.rmtree(self.shardpath)
            else:
                self.shards_exist = True
                self.logger.debug("Shards already exist and `shard_replace` is False")
        if self.shardcache:
            self.logger.debug("Creating shardpath %s"%self.shardpath)
            os.makedirs(self.shardpath, exist_ok=True)
        


        
        # Tokenizer and lengths
        self.tokenizer = kwargs.get("tokenizer")
        self.maxlen = kwargs.get("maxlen")
        self.mlm = kwargs.get("mlm_probability", 0.2)
        self.masking = kwargs.get("masking", mode=="train")

        # This is for memcache and shardcache
        self.getcount = 0
        self.shardcount = 0
        if self.memcache:
            self.refresh_flag = len(dataset)    # Checks when we have processed entire dataset, refreshes mask ids when needed
        if self.shardcache:
            self.refresh_flag = self.shardsize
            # this shuffles the internal shard index, so that we get a random example, instead of in order
            self.shard_internal_shuffle = []
            # This shuffles the shard index, so that each epoch, we load shards in different orders. We will set this up after setting up shards...
            self.shard_shuffle = []

        if int(self.cache) + int(self.memcache) + int(self.shardcache) > 1:
            raise ValueError("Use only one cache type")
        
        self.getter = self.uncachedget  # for each entry, directly convert to features and return
        if self.cache:
            raise NotImplementedError()
            self.getter = self.diskget
        elif self.memcache:
            self.getter = self.memget
        elif self.shardcache:
            self.getter = self.shardget

        # if cache, then we will cache transformations to disk, then load them
        # if memcache, then we will cache to memory and load them
        # if shardcache, we will cache to disk in a sharded fashion, and load shards

        if self.cache or self.memcache:
        # The actual cache-ing
            self.logger.debug("Started mem caching")
            self.input_length_cache = []
            self.logger.debug("Converting to features")
            self.convert_to_features(self.dataset, self.tokenizer, maxlen=self.maxlen)
            self.logger.debug("Masking")
            self.memcached_dataset = self.refresh_mask_ids()

        if self.shardcache:
            self.input_length_cache = []
            if not self.shards_exist:
                self.shardsaveindex = self.sharded_convert_to_features(self.dataset, self.tokenizer, maxlen=self.maxlen)    # save shards and get numshards
            else:
                self.shardsaveindex = len(glob(os.path.join(self.shardpath, "*.pt")))-1   # TODO Bug fix if files do not have consistent numbering
            self.shard_load_index = 0   # self.shardsaveindex is the maximum number of shards
            self.shard_shuffle = list(range(self.shardsaveindex))    # count started from 0
            if self.data_shuffle:
                random.shuffle(self.shard_shuffle)
            self.sharded_dataset = self.load_shard(self.shard_shuffle[self.shard_load_index])
            if self.masking:
                self.sharded_dataset = self.refresh_mask_ids(self.sharded_dataset)  # TODO implement masking (and input_length_cache) for sharding
            self.current_shardsize = len(self.sharded_dataset)
            self.shard_internal_shuffle = list(range(self.current_shardsize))    #count started from 0
            if self.data_shuffle:
                random.shuffle(self.shard_internal_shuffle)



    def shardget(self, idx):
        # get entry from already loaded shardcache. so idx is incremented one by one -- No shuffling in data loader.
        # This is a design choice, and we can't really do anything if user does shuffle + sharding
        self.getcount += 1
        if self.getcount == self.current_shardsize:   # we have exhausted examples in this shard
            self.getcount = 0
            self.shard_load_index += 1                  # increment the shard index that we will load
            if self.shard_load_index == self.shardsaveindex: # we have processed all shards.
                self.shard_load_index = 0
                if self.data_shuffle:
                    random.shuffle(self.shard_shuffle)
            self.sharded_dataset = self.load_shard(self.shard_shuffle[self.shard_load_index])

            if self.masking:
                self.sharded_dataset = self.refresh_mask_ids(self.sharded_dataset)  # TODO implement masking (and input_length_cache) for sharding

            self.current_shardsize = len(self.sharded_dataset)
            self.shard_internal_shuffle = list(range(self.current_shardsize))    #count started from 0
            if self.data_shuffle:
                random.shuffle(self.shard_internal_shuffle)
        #shardindex = idx % self.current_shardsize
        return self.sharded_dataset[self.shard_internal_shuffle[self.getcount]]

    def sharded_convert_to_features(self, dataset, tokenizer, maxlen):
        features = []
        shardsaveindex = 0
        self.input_length_cache = []
        pbar = tqdm(total=int(len(dataset)/self.shardsize)+1)
        for idx, sample in enumerate(dataset):
            tokens = self.tokenizer.tokenize(sample[0])
            if len(tokens) > maxlen - 2:
                tokens = tokens[0:(maxlen - 2)]

            finaltokens = ["[CLS]"]
            token_type_ids = [0]
            for token in tokens:
                finaltokens.append(token)
                token_type_ids.append(0)
            finaltokens.append("[SEP]")
            token_type_ids.append(0)

            input_ids = self.tokenizer.convert_tokens_to_ids(finaltokens)
            attention_mask = [1]*len(input_ids)
            input_len = len(input_ids)
            self.input_length_cache.append(len(input_ids))
            while len(input_ids) < maxlen:
                input_ids.append(0)
                attention_mask.append(0)
                token_type_ids.append(0) 

            assert len(input_ids) == maxlen
            assert len(attention_mask) == maxlen
            assert len(token_type_ids) == maxlen
        
            features.append(
                (input_ids, attention_mask, token_type_ids, input_len, *sample[1:])
            )

            if (idx+1)%self.shardsize == 0: # we need to save this shard.
                # TODO need to handle multiple labels, later...
                all_input_ids = torch.tensor([f[0] for f in features], dtype=torch.long)
                all_attention_mask = torch.tensor([f[1] for f in features], dtype=torch.long)
                all_token_type_ids = torch.tensor([f[2] for f in features], dtype=torch.long)
                all_lens = torch.tensor([f[3] for f in features], dtype=torch.long)
                all_labels = torch.tensor([f[4] for f in features], dtype=torch.long)
                all_masklm = -1*torch.ones(all_input_ids.shape, dtype=torch.long)

                # SAVE HERE
                self.save_shard(shard=TensorDataset(all_input_ids, all_attention_mask, all_token_type_ids, all_masklm, all_lens, all_labels),
                        shard_index = shardsaveindex)
                shardsaveindex += 1
                features = []
                pbar.update(1)
        # final shard...
        if len(features) > 0:
            all_input_ids = torch.tensor([f[0] for f in features], dtype=torch.long)
            all_attention_mask = torch.tensor([f[1] for f in features], dtype=torch.long)
            all_token_type_ids = torch.tensor([f[2] for f in features], dtype=torch.long)
            all_lens = torch.tensor([f[3] for f in features], dtype=torch.long)
            all_labels = torch.tensor([f[4] for f in features], dtype=torch.long)
            all_masklm = -1*torch.ones(all_input_ids.shape, dtype=torch.long)

            # SAVE HERE
            self.save_shard(shard=TensorDataset(all_input_ids, all_attention_mask, all_token_type_ids, all_masklm, all_lens, all_labels),
                    shard_index = shardsaveindex)
            shardsaveindex += 1
            features = []
            pbar.update(1)
        pbar.close()
        return shardsaveindex

    def save_shard(self, shard: TensorDataset, shard_index: int):
        shard_save_path = self.base_shardpath + str(shard_index) + ".pt"
        torch.save(shard, shard_save_path)

    def load_shard(self, shard_index: int):
        shard_save_path = self.base_shardpath + str(shard_index) + ".pt"
        self.logger.debug("Loading shard %s"%shard_save_path)
        return torch.load(shard_save_path)

    def memget(self, idx):
        self.getcount+=1
        if self.getcount > self.refresh_flag and self.masking:
            self.memcached_dataset = self.refresh_mask_ids()
            self.getcount = 0
        return self.memcached_dataset[idx]

    def refresh_mask_ids(self):
        self.logger.debug("Refreshing mask ids")
        if self.masking:
            self.mask_ids = self.build_mask_ids(self.input_length_cache)
            
            all_attention_mask = self.all_attention_mask.clone()
            all_masklm = self.all_masklm.clone()

            for idx in range(self.refresh_flag):
                all_attention_mask[idx][self.mask_ids[idx]] = 0 # Set the masking words to 0, so we do not attend to it during prediction
                all_masklm[idx][self.mask_ids[idx]] = self.all_input_ids[idx][self.mask_ids[idx]] # Set the masking labels for these to the actual word index from all_input_ids

            return TensorDataset(self.all_input_ids, all_attention_mask, self.all_token_type_ids, all_masklm, self.all_lens, self.all_labels)
        else:
            return TensorDataset(self.all_input_ids, self.all_attention_mask, self.all_token_type_ids, self.all_masklm, self.all_lens, self.all_labels)

    def build_mask_ids(self, input_length_cache):
        # for each element, we get a set of indices that are randomly selected...
        # Also, -2 and +1 take care of [cls] and [sep] not being masked
        return [(torch.randperm(inplength-2)+1)[:int(inplength*self.mlm)]  for inplength in input_length_cache]

    
    def sharded_refresh_mask_ids(self):
        pass

    def sharded_build_mask_ids(self, input_length_cache):
        pass


    def convert_to_features(self, dataset, tokenizer, maxlen):
        features = []
        self.input_length_cache = []
        for idx, sample in enumerate(dataset):
            tokens = self.tokenizer.tokenize(sample[0])
            if len(tokens) > maxlen - 2:
                tokens = tokens[0:(maxlen - 2)]

            finaltokens = ["[CLS]"]
            token_type_ids = [0]
            for token in tokens:
                finaltokens.append(token)
                token_type_ids.append(0)
            finaltokens.append("[SEP]")
            token_type_ids.append(0)

            input_ids = self.tokenizer.convert_tokens_to_ids(finaltokens)
            attention_mask = [1]*len(input_ids)
            input_len = len(input_ids)
            self.input_length_cache.append(len(input_ids))
            while len(input_ids) < maxlen:
                input_ids.append(0)
                attention_mask.append(0)
                token_type_ids.append(0) 

            assert len(input_ids) == maxlen
            assert len(attention_mask) == maxlen
            assert len(token_type_ids) == maxlen
        
            features.append(
                (input_ids, attention_mask, token_type_ids, input_len, *sample[1:])
            )

        self.all_input_ids = torch.tensor([f[0] for f in features], dtype=torch.long)
        self.all_attention_mask = torch.tensor([f[1] for f in features], dtype=torch.long)
        self.all_token_type_ids = torch.tensor([f[2] for f in features], dtype=torch.long)
        self.all_lens = torch.tensor([f[3] for f in features], dtype=torch.long)
        self.all_labels = torch.tensor([f[4] for f in features], dtype=torch.long)
        self.all_masklm = -1*torch.ones(self.all_input_ids.shape, dtype=torch.long)
        
    def __len__(self):
        return len(self.dataset)

    def __getitem__(self, idx):
        return self.getter(idx)

    def uncachedget(self, idx):
        return 

        tokens = self.tokenizer.tokenize(self.dataset[idx][0])
        if len(tokens) > maxlen - 2:
            tokens = tokens[0:(maxlen - 2)]

        finaltokens = ["[CLS]"]
        token_type_ids = [0]
        for token in tokens:
            finaltokens.append(token)
            token_type_ids.append(0)
        finaltokens.append("[SEP]")
        token_type_ids.append(0)

        input_ids = self.tokenizer.convert_tokens_to_ids(finaltokens)
        attention_mask = [1]*len(input_ids)
        input_len = len(input_ids)

        while len(input_ids) < max_seq_length:
            input_ids.append(0)
            attention_mask.append(0)
            token_type_ids.append(0) 

        assert len(input_ids) == max_seq_length
        assert len(attention_mask) == max_seq_length
        assert len(token_type_ids) == max_seq_length
        
        return (input_ids, attention_mask, token_type_ids, input_len, self.dataset[idx][1])


from ednaml.utils.LabelMetadata import LabelMetadata
from ednaml.generators import TextGenerator
class AlbertGenerator(TextGenerator):
  # input includes tokenizer for build...

  # Set it up such that, given a crawler, create a dataset from it.
  # Then create a cached batch
  # From cached batch, yield batches until it is empty

  def build_transforms(self, transform, mode, **kwargs):  #<-- generator kwargs:
    from ednaml.utils import locate_class
    print("Building Transforms")
    tokenizer = kwargs.get("tokenizer", "AlbertFullTokenizer")
    self.tokenizer = locate_class(package="ednaml", subpackage="utils", classpackage=tokenizer, classfile="tokenizers")
    self.tokenizer = self.tokenizer(**kwargs) # vocab_file, do_lower_case, spm_model_file

  
  def buildDataset(self, crawler, mode, transform, **kwargs): #<-- dataset args:
    print("Building Dataset")
    return AlbertDataset(self.logger, crawler.metadata[mode]["crawl"], mode, tokenizer = self.tokenizer, **kwargs) # needs maxlen, memcache, mlm_probability

  def buildDataLoader(self, dataset, mode, batch_size, **kwargs):
    print("Building Dataloader")
    return torch.utils.data.DataLoader(dataset, batch_size=batch_size*self.gpus,
                                        shuffle=kwargs.get("shuffle", True), num_workers = self.workers, 
                                       collate_fn=self.collate_fn)

  def getNumEntities(self, crawler, mode, **kwargs):  #<-- dataset args
    label_dict = {
        item: {"classes": crawler.metadata[mode]["classes"][item]}
        for item in kwargs.get("classificationclass", ["color"])
    }
    return LabelMetadata(label_dict=label_dict)

  def collate_fn(self, batch):
    all_input_ids, all_attention_mask, all_token_type_ids, all_masklm, all_lens, all_labels  = map(torch.stack, zip(*batch))
    max_len = max(all_lens).item()
    all_input_ids = all_input_ids[:, :max_len]
    all_attention_mask = all_attention_mask[:, :max_len]
    all_token_type_ids = all_token_type_ids[:, :max_len]
    all_masklm = all_masklm[:, :max_len]

    return all_input_ids, all_attention_mask, all_token_type_ids, all_masklm, all_labels







import tqdm, json
from sklearn.metrics import f1_score
import shutil
import os
import torch
import numpy as np
from ednaml.trainer import BaseTrainer


class ClassificationTrainer(BaseTrainer):
    def init_setup(self, **kwargs):
        self.softaccuracy = []

    # Steps through a batch of data
    def step(self, batch): 
        batch_kwargs = {}
        (
            img,
            batch_kwargs["labels"],
        ) = batch  
        # logits, features, labels
        batch_kwargs["logits"], batch_kwargs["features"], _ = self.model(img)
        batch_kwargs["epoch"] = self.global_epoch  # For CompactContrastiveLoss

        # TODO fix this with info about how many output are in the model...from the config file!!!!!
        loss = {loss_name: None for loss_name in self.loss_fn}
        for lossname in self.loss_fn:
            loss[lossname] = self.loss_fn[lossname](**batch_kwargs)
        # if self.fp16 and self.apex is not None:
        #    with self.apex.amp.scale_loss(loss, self.optimizer) as scaled_loss:
        #        scaled_loss.backward()
        # else:
        #    loss.backward()
        lossbackward = sum(loss.values())

        for _, lossname in enumerate(self.loss_fn):
            self.losses[lossname].append(loss[lossname].cpu().item())

        if batch_kwargs["logits"] is not None:
            softmax_accuracy = (
                (batch_kwargs["logits"].max(1)[1] == batch_kwargs["labels"])
                .float()
                .mean()
            )
            self.softaccuracy.append(softmax_accuracy.cpu().item())
        else:
            self.softaccuracy.append(0)

        return lossbackward

    def evaluate_impl(self):
        self.model.eval()
        features, logits, labels = [], [], []
        with torch.no_grad():
            for batch in tqdm.tqdm(
                self.test_loader, total=len(self.test_loader), leave=False
            ):
                data, label = batch
                data = data.to(self.device)
                logit, feature, _ = self.model(data)
                feature = feature.detach().cpu()
                logit = logit.detach().cpu()
                features.append(feature)
                logits.append(logit)
                labels.append(label)

        features, logits, labels = (
            torch.cat(features, dim=0),
            torch.cat(logits, dim=0),
            torch.cat(labels, dim=0),
        )
        # Now we compute the loss...
        self.logger.info("Obtained features, validation in progress")
        # for evaluation...

        logit_labels = torch.argmax(logits, dim=1)
        accuracy = (logit_labels == labels).sum().float() / float(
            labels.size(0)
        )
        micro_fscore = np.mean(f1_score(labels, logit_labels, average="micro"))
        weighted_fscore = np.mean(
            f1_score(labels, logit_labels, average="weighted")
        )
        self.logger.info("Accuracy: {:.3%}".format(accuracy))
        self.logger.info("Micro F-score: {:.3f}".format(micro_fscore))
        self.logger.info("Weighted F-score: {:.3f}".format(weighted_fscore))
        return logit_labels, labels, logits

    def saveMetadata(
        self,
    ):
        self.logger.info("Saving model metadata")
        jMetadata = json.dumps(self.metadata)
        metafile = "metadata.json"
        localmetafile = os.path.join(self.save_directory, metafile)
        if self.save_backup:
            backupmetafile = os.path.join(self.backup_directory, metafile)
        if not os.path.exists(localmetafile):
            with open(localmetafile, "w") as localmetaobj:
                localmetaobj.write(jMetadata)
        self.logger.info("Backing up metadata")
        if self.save_backup:
            shutil.copy2(localmetafile, backupmetafile)
        self.logger.info("Finished metadata backup")



from logging import Logger
from os import PathLike
from torch import TensorType, nn
import torch
from typing import Any, List, Dict, Tuple, Type
from ednaml.plugins import ModelPlugin
from ednaml.utils.LabelMetadata import LabelMetadata


class ModelAbstract(nn.Module):
    """ModelAbstract is the base class for an Edna Model used in EdnaML, EdnaDeploy, and any other Edna frameworks. ModelAbstract allows design of practically any neural-network based trainable model by inheriting from torch.nn.

    ModelAbstract allows users to define a model by following specific recipes. While reducing some flexibibility, the end result is to allow for better interoperability and extensibility.

    Attributes:
        model_name (str): The name for this model. Useful during debugging.
        model_arch (TODO): TODO
        number_outputs (int): The number of outputs for this model
        output_classnames (List[str]): The name for each output for this model. This can be set manually or programatically.
        output_dimensions (List[int]): The dimensions for each output for this model. This can be set manually or programatically.
        secondary_outputs (TODO): TODO
        model_base (str): Name for the core model architecture, of this is a modular construction. Useful during debugging.
        weights (str): Path to weights file for this model
        normalization (str): Name for type of normalization used in this model. Provided here since it is a common model characteristic.
        metadata (LabelMetadata): A LabelMetadata object for this model that provides information on the training input and mopdel output. 
        parameter_groups (Dict[str, nn.Module]): A named dictionary of a ModelAbstracts sub-components. For example, a GAN can have encoder, decoder, and discriminator parameter groups, if they are trained separately.

    Methods:
        _type_: _description_
    """
    model_name: str = "ModelAbstract"
    model_arch: str = None
    number_outputs: int = 1
    output_classnames: List[str] = ["out1"]
    output_dimensions: List[int] = [512]
    secondary_outputs: List[Any] = []

    model_base: str
    weights: str
    normalization: str
    metadata: LabelMetadata
    parameter_groups: Dict[str, nn.Module]

    plugins: Dict[str,ModelPlugin] = {}
    plugin_count: int = 0
    has_plugins: bool = False
    plugin_hook: str = "always"

    _logger: Logger = None

    def __init__(
        self,
        base: str = None,
        weights: PathLike = None,
        metadata: LabelMetadata = None,
        normalization: str = None,
        parameter_groups: List[str] = None,
        **kwargs
    ):
        super().__init__()
        self.metadata: LabelMetadata = metadata
        self.model_base = base
        self.weights = weights
        self.normalization = normalization
        self.parameter_groups = {}
        self.inferencing = False
        

        self.model_attributes_setup(**kwargs)
        self.model_setup(**kwargs)
        self.parameter_groups_setup(parameter_groups)

    def set_plugin_hooks(self, plugin_hook: str):   # always | warmup | activated
        self.plugin_hook = plugin_hook

    def model_attributes_setup(self, **kwargs):
        raise NotImplementedError()

    def model_setup(self, **kwargs):
        raise NotImplementedError()

    def parameter_groups_setup(self, parameter_groups: List[str]):
        self.parameter_groups[parameter_groups[0]] = "self"

    def weights_init_kaiming(self, m):
        classname = m.__class__.__name__
        if classname.find("Linear") != -1:
            nn.init.kaiming_normal_(m.weight, a=0, mode="fan_out")
            nn.init.constant_(m.bias, 0.0)
        elif classname.find("Conv") != -1:
            nn.init.kaiming_normal_(m.weight, a=0, mode="fan_in")
            if m.bias is not None:
                nn.init.constant_(m.bias, 0.0)
        elif classname.find("BatchNorm") != -1:
            if m.affine:
                nn.init.constant_(m.weight, 1.0)
                nn.init.constant_(m.bias, 0.0)
        elif classname.find("GroupNorm") != -1:
            if m.affine:
                nn.init.constant_(m.weight, 1.0)
                nn.init.constant_(m.bias, 0.0)
        elif classname.find("LayerNorm") != -1:
            # if m.affine:
            #    nn.init.constant_(m.weight, 1.0)
            #    nn.init.constant_(m.bias, 0.0)
            pass
        elif classname.find("InstanceNorm") != -1:
            if m.affine:
                nn.init.constant_(m.weight, 1.0)
                nn.init.constant_(m.bias, 0.0)

    def weights_init_softmax(self, m):
        """Initialize linear weights to standard normal. Mean 0. Standard Deviation 0.001"""
        classname = m.__class__.__name__
        if classname.find("Linear") != -1:
            nn.init.normal_(m.weight, std=0.001)
            if m.bias:
                nn.init.constant_(m.bias, 0.0)

    def partial_load(self, weights_path):
        params = torch.load(weights_path)
        for _key in params:
            if (
                _key not in self.state_dict().keys()
                or params[_key].shape != self.state_dict()[_key].shape
            ):
                continue
            self.state_dict()[_key].copy_(params[_key])

    def forward(self, x, labels=None, **kwargs):    # TODO Labels for the plugins...?????
        if self.training and self.inferencing:
            raise ValueError(
                "Cannot inference and train at the same time! Call"
                " deinference() first, before train()"
            )

        # TODO, we neede a pre-forward hook, a forward hook, and a post foward hook
        if self.has_plugins:
            x, kwargs, secondary_output_queue_pre = self.pre_forward_hook(x, **kwargs)
        feature_logits, features, secondary_outputs = self.forward_impl(
            x, **kwargs
        )
        if self.has_plugins:
            feature_logits, features, secondary_outputs, secondary_output_queue_post = self.post_forward_hook(x, feature_logits, features, secondary_outputs, **kwargs)

        # TODO deal with secondary_output_queues
        if self.has_plugins:
            secondary_outputs = (secondary_outputs, secondary_output_queue_pre, secondary_output_queue_post)
        return feature_logits, features, secondary_outputs

    def foward_impl(self, x, **kwargs) -> Tuple[TensorType,TensorType,List[Any]]:   # Return Logits, Features, Secondary Outputs

        raise NotImplementedError()

    def getModelName(self):
        return self.model_name

    def getModelBase(self):
        return self.model_base

    def getModelArch(self):
        return self.model_arch

    def getParameterGroup(self, key: str) -> nn.Module:
        if (
            self.parameter_groups[key] == "self"
        ):  # avoid recursion bug, I think?
            return self
        else:
            return self.parameter_groups[key]

    def inference(self):
        self.eval()
        self.inferencing = True

    def deinference(self):
        self.inferencing = False

    """
    def train(self, mode: bool = True):
        self.inferencing = False
        return super().train(mode)

    def eval(self):
        self.inferencing = False
        return super().eval()
    """

    def convertForInference(self) -> "ModelAbstract":
        raise NotImplementedError

    #-------------------------------------------------------------------------------------------
    # Model Plugins and Hooks architecture
    def loadPlugins(self, plugin_path: PathLike, ignore_plugins: List[str] = []):
        plugin_dict = torch.load(plugin_path)
        for plugin_name in plugin_dict:
            if plugin_name in self.plugins:
                if plugin_name not in ignore_plugins:
                    self.plugins[plugin_name].load(plugin_dict[plugin_name])
                    self._logger.info("Loading plugin with name %s"%plugin_name)
                else:
                    self._logger.info("Ignoring plugin with name %s"%plugin_name)
            else:
                self._logger.info("No plugin exists for name %s"%plugin_name)

    def savePlugins(self):
        save_dict = {}
        for plugin_name in self.plugins:
            save_dict[plugin_name] = self.plugins[plugin_name].save()
            

        return save_dict

    def addPlugin(self, plugin: Type[ModelPlugin], plugin_name: str = None, plugin_kwargs: Dict[str, Any] = {}):
        if plugin_name is None:
            plugin_name = plugin.name
        if plugin_name == "ModelPlugin":
            # TODO NOTE we are badly changing the ModelPlugin name to the class name
            raise ValueError("Potentially no actual plugin passed!")

        if plugin_name in self.plugins:
            raise KeyError("`plugin_name` %s already exists in self.plugins:  "%plugin_name)
        else:
            self._logger.info("Added plugin %s"%plugin_name)
            self.plugins[plugin_name] = plugin(**plugin_kwargs)
            self.plugins[plugin_name]._logger = self._logger
        
        self.plugin_count = len(self.plugins)
        self.has_plugins = self.plugin_count > 0

    def set_usable_plugins(self):
        if self.plugin_hook == "always":
            self.usable_plugins = [item for item in self.plugins]
        elif self.plugin_hook == "warmup":
            self.usable_plugins = [item for item in self.plugins if not self.plugins[item].activated]
        elif self.plugin_hook == "activated":
            self.usable_plugins = [item for item in self.plugins if self.plugins[item].activated]
        else:
            raise ValueError("Unknown value for `plugin_hook`: %s"%self.plugin_hook)

    def pre_epoch_hook(self, epoch: int = 0):
        # Here, we set up which plugins are activated and can be used, based on cfg...PLUGIN.HOOK
        # we do this check before, and after...
        self.set_usable_plugins()   # dealing with a plugin-load
        for plugin in self.usable_plugins:
            self.plugins[plugin].pre_epoch(model = self, epoch=epoch)
        self.set_usable_plugins()   # post plugin usage
    def post_epoch_hook(self, epoch: int = 0):
        # Here, we set up which plugins are activated and can be used, based on cfg...PLUGIN.HOOK
        # we do this check before, and after...
        #self.set_usable_plugins()  # unneeded
        for plugin in self.usable_plugins:
            self.plugins[plugin].post_epoch(model = self, epoch=epoch)
        self.set_usable_plugins()   # post plugin usage

    def pre_forward_hook(self, x, **kwargs):
        # For example, L-Score will perturb the input, then itself call self.forward_impl with the perturbed input (i.e. itself very 
        # hacky under our own framework), then provide results to secondary_outputs
        secondaries: Dict[str, Any] = {}
        for plugin in self.usable_plugins:
            x, kwargs, secondary_output_pre = self.plugins[plugin].pre_forward(x, **kwargs)
            secondaries[plugin] = secondary_output_pre
        return x, kwargs, secondaries

    def post_forward_hook(self, x, feature_logits, features, secondary_outputs, **kwargs):
        # For ModelPlugins.
        # For example, KMP, after forward pass, provides distance to nearest proxy as well as nearest proxy in the secondary_outputs
        secondaries: Dict[str, Any] = {}
        for plugin in self.usable_plugins:
            feature_logits, features, secondary_outputs, kwargs, secondary_output_post = self.plugins[plugin].post_forward(x,feature_logits, features, secondary_outputs,self, **kwargs)
            secondaries[plugin] = secondary_output_post
        return feature_logits, features, secondary_outputs, secondaries



import torch


def pairwise_distance(a, squared=False, eps=1e-16):
    """Computes the pairwise distance matrix with numerical stability."""
    operand = a.pow(2).sum(dim=1, keepdim=True).expand(a.size(0), -1)
    operandT = torch.t(a).pow(2).sum(dim=0, keepdim=True).expand(a.size(0), -1)
    pairwise_distances_squared = torch.add(operand, operandT) - 2 * (
        torch.mm(a, torch.t(a))
    )
    pairwise_distances_squared = torch.clamp(
        pairwise_distances_squared, min=0.0
    )
    error_mask = torch.le(pairwise_distances_squared, 0.0)

    if squared:
        pairwise_distances = pairwise_distances_squared
    else:
        pairwise_distances = torch.sqrt(
            pairwise_distances_squared + error_mask.float() * eps
        )

    # Undo conditionally adding 1e-16.
    pairwise_distances = torch.mul(
        pairwise_distances, (error_mask == False).float()
    )

    # Explicitly set diagonals to zero.
    mask_offdiagonals = 1 - torch.eye(
        *pairwise_distances.size(), device=pairwise_distances.device
    )
    pairwise_distances = torch.mul(pairwise_distances, mask_offdiagonals)
    return pairwise_distances


def o_metric(x_matrix, y_matrix, directional = False):
    """ Computes the point-proximity O metric from 
    "Analysing patterns of spatial and niche overlap among species at multiple resolutions", 
    by Marcel Cardillo and Dan Warren, published in Global Ecology and BioGeography, 2016.


    Given two matrices, an `x_matrix` of M x n, and a `y_matrix`
    of P x n, both in n-dimensional space of M and P points, respectively,
    this function computes the point-proximity O metric for the two datasets.

    """
    con_x = torch.cdist(x_matrix.unsqueeze(0), x_matrix.unsqueeze(0))[0]
    if not directional:
      con_y = torch.cdist(y_matrix.unsqueeze(0), y_matrix.unsqueeze(0))[0]
    hetero = torch.cdist(x_matrix.unsqueeze(0), y_matrix.unsqueeze(0))[0]

    con_x.fill_diagonal_(torch.max(con_x))
    if not directional:
      con_y.fill_diagonal_(torch.max(con_y))

    con_x = torch.min(con_x, dim=1)[0]
    if not directional:
      con_y = torch.min(con_y, dim=1)[0]

    hetero_x = torch.min(hetero, dim=1)[0]
    if not directional:
      hetero_y = torch.min(hetero.T, dim=1)[0]

    o_x = con_x / hetero_x
    if not directional:
      o_y = con_y / hetero_y

    p_x = torch.sum((o_x>1).float()) / o_x.shape[0]
    if not directional:
      p_y = torch.sum((o_y>1).float()) / o_y.shape[0]

    if directional:
      o_val = p_x
    else:
      o_val = (p_x + p_y) / 2
    return o_val


#!/usr/bin/env python3
"""Generators yields batches to models during training and testing. 
Requirs a crawler to provide a list of paths with labels.
Generator then loads a random subset, performs transformations, 
and yields these to model when requested.

List of supported generators:

    - SequencedGenerator: Yields a triplet batch for re-id
    - TripletGenerator: Yields a triplet batch for re-id
    - ClassedGenerator:
    - Cars196Generator:
    - ClassificationGenerator:
    - CoLabelIntegratedDatasetGenerator: 
    - CoLabelDeployGenerator:
    - KnowledgeIntegratedGenerator: 
"""

from logging import Logger
from typing import Any, Dict, List
import torchvision.transforms as T
from torch.utils.data import Dataset as TorchDataset
from torch.utils.data import DataLoader as TorchDataLoader
from ednaml.utils.LabelMetadata import LabelMetadata


class Generator:
    num_entities: LabelMetadata
    dataloader: TorchDataLoader
    dataset: TorchDataset
    gpus: int
    mode: str
    training_mode: bool

    def __init__(
        self,
        logger: Logger,
        gpus: int = 1,
        transforms: Dict[str, Any] = {},
        mode: str = "train",
        **kwargs
    ):
        """Initialize the generator"""
        self.logger = logger
        self.gpus = max(1, gpus)
        self.dataloader = None
        self.transforms = transforms
        self.mode = mode
        self.training_mode = self.isTrainingMode()
        self.transformer = self.build_transforms(
            self.transforms, self.mode, **kwargs
        )
        self.buildGeneratorAttributes(**kwargs)	
    
    
    def buildGeneratorAttributes(self, **kwargs):	
        pass
    
    def build_transforms(self, transforms: Dict[str, Any], mode, **kwargs):
        return None

    def isTrainingMode(self):
        return self.mode == "train"

    def build(self, datacrawler, batch_size, workers, **kwargs):
        """This should generate a TorchDataset and associated DataLoader to yield batches.

        Raises:
            NotImplementedError: _description_
        """

        self.workers = workers * self.gpus

        self.dataset = self.buildDataset(
            datacrawler, self.mode, self.transformer, **kwargs
        )
        self.dataloader = self.buildDataLoader(
            self.dataset, self.mode, batch_size=batch_size, **kwargs
        )
        self.num_entities = self.getNumEntities(
            datacrawler, self.mode, **kwargs
        )

    def buildDataset(
        self, datacrawler, mode: str, transform: List[object], **kwargs
    ) -> TorchDataset:
        """Given the datacrawler with all the data, and the mode (could be
        any user-defined mode such as 'train', 'test', 'zsl', 'gzsl', etc), as
        as well the transform, return a TorchDataset

        Args:
            datacrawler (_type_): _description_
            mode (_type_): _description_
            transform (_type_): _description_

        Raises:
            NotImplementedError: _description_
        """
        return None

    def buildDataLoader(self, dataset, mode, batch_size, **kwargs):
        """Given a Torch Dataset, build a TorchDataloader to return a tensor, possibly with a collate function

        Args:
            dataset (_type_): _description_
            batch_size (_type_): _description_

        Raises:
            NotImplementedError: _description_
        """

        return None

    def getNumEntities(self, datacrawler, mode, **kwargs) -> LabelMetadata:
        """Return the number of classes in the overall label scheme
        Raises:
            NotImplementedError: _description_
        """

        return None


from ednaml.generators.ImageGenerator import ImageGenerator
from ednaml.generators.TextGenerator import TextGenerator

from ednaml.generators.ClassificationGenerator import ClassificationGenerator
from ednaml.generators.MultiClassificationGenerator import (
    MultiClassificationGenerator,
)
from ednaml.generators.TorchvisionGeneratorWrapper import (
    TorchvisionGeneratorWrapper,
)
from ednaml.generators.HFGenerator import HFGenerator



import unicodedata
from ednaml.utils.tokenizers import (
    whitespace_tokenize,
    load_vocab,
    encode_pieces,
    convert_by_vocab,
    _is_control,
    _is_punctuation,
    _is_whitespace,
)
import sentencepiece


class AlbertFullTokenizer(object):
    """Runs end-to-end tokenziation."""

    def __init__(
        self, vocab_file, do_lower_case=True, spm_model_file=None, **kwargs
    ):
        self.vocab = None
        self.sp_model = None
        if spm_model_file:
            self.sp_model = sentencepiece.SentencePieceProcessor()
            # logger.info("loading sentence piece model")
            self.sp_model.Load(spm_model_file)
            # # Note(mingdachen): For the purpose of consisent API, we are
            # # generating a vocabulary for the sentence piece tokenizer.
            self.vocab = {
                self.sp_model.IdToPiece(i): i
                for i in range(self.sp_model.GetPieceSize())
            }
        else:
            print("load vocab")
            self.vocab = load_vocab(vocab_file)
            print("load token")
            self.basic_tokenizer = AlbertBasicTokenizer(
                do_lower_case=do_lower_case
            )
            self.wordpiece_tokenizer = AlbertWordpieceTokenizer(
                vocab=self.vocab,
                unk_token="[UNK]",
                max_input_chars_per_word=100,
            )
        self.inv_vocab = {v: k for k, v in self.vocab.items()}

    def tokenize(self, text):
        if self.sp_model:
            split_tokens = encode_pieces(
                self.sp_model, text, return_unicode=False
            )
        else:
            split_tokens = []
            for token in self.basic_tokenizer.tokenize(text):
                for sub_token in self.wordpiece_tokenizer.tokenize(token):
                    split_tokens.append(sub_token)

        return split_tokens

    def convert_tokens_to_ids(self, tokens):
        if self.sp_model:
            return [self.sp_model.PieceToId(token) for token in tokens]
        else:
            return convert_by_vocab(self.vocab, tokens)

    def convert_ids_to_tokens(self, ids):
        if self.sp_model:
            return [self.sp_model.IdToPiece(id_) for id_ in ids]
        else:
            return convert_by_vocab(self.inv_vocab, ids)


class AlbertBasicTokenizer(object):
    """Runs basic tokenization (punctuation splitting, lower casing, etc.)."""

    def __init__(self, do_lower_case=True, **kwargs):
        """Constructs a BasicTokenizer.
        Args:
          do_lower_case: Whether to lower case the input.
        """
        self.do_lower_case = do_lower_case

    def tokenize(self, text):
        """Tokenizes a piece of text."""
        text = self._clean_text(text)

        # This was added on November 1st, 2018 for the multilingual and Chinese
        # models. This is also applied to the English models now, but it doesn't
        # matter since the English models were not trained on any Chinese data
        # and generally don't have any Chinese data in them (there are Chinese
        # characters in the vocabulary because Wikipedia does have some Chinese
        # words in the English Wikipedia.).
        text = self._tokenize_chinese_chars(text)
        orig_tokens = whitespace_tokenize(text)
        split_tokens = []
        for token in orig_tokens:
            if self.do_lower_case:
                token = token.lower()
                token = self._run_strip_accents(token)
            split_tokens.extend(self._run_split_on_punc(token))
        output_tokens = whitespace_tokenize(" ".join(split_tokens))
        return output_tokens

    def _run_strip_accents(self, text):
        """Strips accents from a piece of text."""
        text = unicodedata.normalize("NFD", text)
        output = []
        for char in text:
            cat = unicodedata.category(char)
            if cat == "Mn":
                continue
            output.append(char)
        return "".join(output)

    def _run_split_on_punc(self, text):
        """Splits punctuation on a piece of text."""
        chars = list(text)
        i = 0
        start_new_word = True
        output = []
        while i < len(chars):
            char = chars[i]
            if _is_punctuation(char):
                output.append([char])
                start_new_word = True
            else:
                if start_new_word:
                    output.append([])
                start_new_word = False
                output[-1].append(char)
            i += 1

        return ["".join(x) for x in output]

    def _tokenize_chinese_chars(self, text):
        """Adds whitespace around any CJK character."""
        output = []
        for char in text:
            cp = ord(char)
            if self._is_chinese_char(cp):
                output.append(" ")
                output.append(char)
                output.append(" ")
            else:
                output.append(char)
        return "".join(output)

    def _is_chinese_char(self, cp):
        """Checks whether CP is the codepoint of a CJK character."""
        # This defines a "chinese character" as anything in the CJK Unicode block:
        #   https://en.wikipedia.org/wiki/CJK_Unified_Ideographs_(Unicode_block)
        #
        # Note that the CJK Unicode block is NOT all Japanese and Korean characters,
        # despite its name. The modern Korean Hangul alphabet is a different block,
        # as is Japanese Hiragana and Katakana. Those alphabets are used to write
        # space-separated words, so they are not treated specially and handled
        # like the all of the other languages.
        if (
            (cp >= 0x4E00 and cp <= 0x9FFF)
            or (cp >= 0x3400 and cp <= 0x4DBF)  #
            or (cp >= 0x20000 and cp <= 0x2A6DF)  #
            or (cp >= 0x2A700 and cp <= 0x2B73F)  #
            or (cp >= 0x2B740 and cp <= 0x2B81F)  #
            or (cp >= 0x2B820 and cp <= 0x2CEAF)  #
            or (cp >= 0xF900 and cp <= 0xFAFF)
            or (cp >= 0x2F800 and cp <= 0x2FA1F)  #
        ):  #
            return True

        return False

    def _clean_text(self, text):
        """Performs invalid character removal and whitespace cleanup on text."""
        output = []
        for char in text:
            cp = ord(char)
            if cp == 0 or cp == 0xFFFD or _is_control(char):
                continue
            if _is_whitespace(char):
                output.append(" ")
            else:
                output.append(char)
        return "".join(output)


class AlbertWordpieceTokenizer(object):
    """Runs WordPiece tokenization."""

    def __init__(self, vocab, unk_token, max_input_chars_per_word=100):
        self.vocab = vocab
        self.unk_token = unk_token
        self.max_input_chars_per_word = max_input_chars_per_word

    def tokenize(self, text, **kwargs):
        """Tokenizes a piece of text into its word pieces.
        This uses a greedy longest-match-first algorithm to perform tokenization
        using the given vocabulary.
        For example:
          input = "unaffable"
          output = ["un", "##aff", "##able"]
        Args:
          text: A single token or whitespace separated tokens. This should have
            already been passed through `BasicTokenizer`.
        Returns:
          A list of wordpiece tokens.
        """

        output_tokens = []
        for token in whitespace_tokenize(text):
            chars = list(token)
            if len(chars) > self.max_input_chars_per_word:
                output_tokens.append(self.unk_token)
                continue

            is_bad = False
            start = 0
            sub_tokens = []
            while start < len(chars):
                end = len(chars)
                cur_substr = None
                while start < end:
                    substr = "".join(chars[start:end])
                    if start > 0:
                        substr = "##" + substr
                    if substr in self.vocab:
                        cur_substr = substr
                        break
                    end -= 1
                if cur_substr is None:
                    is_bad = True
                    break
                sub_tokens.append(cur_substr)
                start = end

            if is_bad:
                output_tokens.append(self.unk_token)
            else:
                output_tokens.extend(sub_tokens)
        return output_tokens



from ednaml.config import BaseConfig
from ednaml.config.ConfigDefaults import ConfigDefaults


class LoggingConfig(BaseConfig):
    STEP_VERBOSE: int

    def __init__(self, logging_dict, defaults: ConfigDefaults):
        self.STEP_VERBOSE = logging_dict.get(
            "STEP_VERBOSE", defaults.STEP_VERBOSE
        )
        self.INPUT_SIZE = logging_dict.get(
            "INPUT_SIZE", defaults.INPUT_SIZE
        )



import os
from typing import Dict


class Crawler:
    """The base crawler class. TODO"""

    classes: Dict[str, int]

    def __init__(self, logger):
        self.classes = {}

    def __verify(self, folder):
        if not os.path.exists(folder):
            raise IOError(
                "Folder {data_folder} does not exist".format(data_folder=folder)
            )
        else:
            self.logger.info("Found {data_folder}".format(data_folder=folder))






from abc import ABC
import json, copy


class BaseConfig(ABC):
    def getVars(self):
        return vars(self)

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __repr__(self):
        return self.to_json_string()

    def to_dict(self):
        """Serializes this instance to a Python dictionary."""
        output = copy.deepcopy(self.__dict__)
        return output

    def to_json_string(self):
        """Serializes this instance to a JSON string."""
        return json.dumps(self.to_dict(), indent=2, sort_keys=True, default=lambda o: o.__dict__)

    def to_json_file(self, json_file_path):
        """Save this instance to a json file."""
        with open(json_file_path, "w", encoding="utf-8") as writer:
            writer.write(self.to_json_string())



import copy
from logging import Logger
from torch import nn
class ModelPlugin():
    name: str = "ModelPlugin"
    _logger: Logger = None
    activated: bool = False
    def __init__(self, **kwargs):
        #super().__init__()
        self.build_plugin(**kwargs)
        self.build_params(**kwargs)
        
        
    def build_plugin(self, **kwargs):
        """Sets up plugin-specific variables
        """
        pass

    def build_params(self, **kwargs):
        """Sets up plugin operational parameters
        """
        pass

    def pre_forward(self, x, **kwargs):
        """Called before the forward pass of a ModelAbstract, to modify the inputs if needed and to add to the secondary outputs

        Args:
            x (_type_): _description_
        """
        return x, kwargs, {}

    def post_forward(self, x, feature_logits, features, secondary_outputs, model, **kwargs):
        """Called after the forward pass of a ModelAbstract, to use Plugin operations parameters to add to secondary outputs and adjust the outputs

        Args:
            x (_type_): _description_
            feature_logits (_type_): _description_
            features (_type_): _description_
            secondary_outputs (_type_): _description_
        """
        return feature_logits, features, secondary_outputs, kwargs, {}

    def pre_epoch(self, model: 'ModelAbstract', epoch: int = 0, **kwargs):
        """Hook that is executed during the model training by BaseTrainer. Occurs before an epoch starts

        Args:
            epoch (int, optional): _description_. Defaults to 0.
        """
        pass

    def post_epoch(self, model: 'ModelAbstract', epoch: int = 0, **kwargs):
        """Hook that is executed during model training by BaseTrainer. Occurs at the end of an epoch.

        Args:
            epoch (int, optional): _description_. Defaults to 0.
        """
        pass

    def save(self):
        return copy.deepcopy({item:self.__dict__[item] for item in self.__dict__.keys() if item[0] != "_"})
        #builtin_keys = [item for item in sdict.keys() if item[0] == "_"]
        #[sdict.pop(key) for key in builtin_keys]
        #return sdict
        # returns an object of itself for saving

    def load(self, save_dict_or_path):
        # TODO adjust this so plugin can be loaded from a file, instead of directly passing parameters...
        self.__dict__.update(save_dict_or_path) # probably need a better way to do this, i.e. iterate through dict, and keep what is ther, warn about extra keys, etc...


import math, torch
from torch import nn
class KMeansProxy(nn.Module):

    def __init__(self, proxy_factor, num_classes, embedding_dimensions):
        super().__init__()
        self.proxy_factor = proxy_factor
        self.num_classes = num_classes
        self.clusters = self.proxy_factor * self.num_classes
        self.embedding_dimensions = embedding_dimensions
        self.lr = 1e-2
        self.proxies = nn.Parameter(torch.zeros(self.clusters, self.embedding_dimensions))
        self.labels = torch.zeros(self.clusters, self.num_classes)

        self._proxiesFlag = False
        self._optimizer = torch.optim.Adam(self.parameters(), lr=self.lr)
        self.cost_window = 25
        self.costs = []
        self.loss = nn.MSELoss()

    def _initialize(self, x):
     
        samples_per_cluster = math.ceil(float(x.size(0))) / self.clusters
        assignment = list(range(0, self.clusters)) * int(samples_per_cluster)
        length_difference = len(assignment) - x.size(0)
        if length_difference < 0:
          assignment = assignment + assignment[length_difference:]
        else:
            assignment = assignment[:len(assignment)-length_difference]
        randomized_order = torch.randperm(len(assignment))
        randomized_assignment = torch.LongTensor(assignment)[randomized_order]
        
        batch_dim = 0
        
        for i in range(self.clusters):
            self.proxies.data[i] += x[randomized_assignment == i].mean(batch_dim)
        
        return randomized_assignment

    def _assign(self, x):
        return (
            ((x[:,None] - self.proxies) ** 2).mean(2).argmin(1)
        )

    def forward(self, x):
        if self.training:
            self.fit(x)
        return x, self.proxies[self._assign(x)], self.labels[self._assign(x)]

    def transform(self, x):
        return self._assign(x)

    def _update(self, x, cluster):
        self._optimizer.zero_grad()
        means = self.proxies[cluster]
        cur_cost = self.loss(x, means)
        cur_cost.backward()
        self._optimizer.step()
        return cur_cost.item()
    
    def fit(self, x):
        if not self._proxiesFlag: 
            self._initialize(x)
            self._proxiesFlag = True
        clusters = self._assign(x)
        cur_cost = self._update(x, clusters)
            #costs.append(cur_cost)
            
    def labelUpdate(self, model, return_tuple_index = None):
        for idx, x in enumerate(self.proxies):
            return_value = model(x)
            self.labels[idx,:] = return_value[return_tuple_index]


from typing import Type, Union, List
from ednaml.core import EdnaML
import logging
from ednaml.crawlers import Crawler
from ednaml.datareaders import DataReader
from ednaml.generators import Generator

from ednaml.deploy.BaseDeploy import BaseDeploy
from ednaml.utils import locate_class

class EdnaDeploy(EdnaML):
    def __init__(
        self,
        config: str = "",
        deploy: Union[List[str], str] = "",
        mode: str = "test",
        weights: str = None,
        logger: logging.Logger = None,
        verbose: int = 2,
        **kwargs
    ):
        
        super().__init__([config,deploy], mode, weights, logger, verbose,**kwargs)

        self.decorator_reference["deployment"] = self.addDeploymentClass
        self.decorator_reference.pop("trainer") # We do not need trainer in a Deployment
        self.dataloader_mode = kwargs.get("dataloader_mode", self.mode)

    def apply(self, **kwargs):
        """Applies the internal configuration for EdnaDeploy"""
        self.printConfiguration()
        self.downloadModelWeights()
        self.setPreviousStop()

        self.buildDataloaders()

        self.buildModel()
        self.loadWeights()
        self.getModelSummary(**kwargs) 

        self.buildDeployment()

        self.resetQueues()


    def buildDataloaders(self):
        """Sets up the datareader classes and builds the train and test dataloaders"""

        data_reader: Type[DataReader] = locate_class(
            package="ednaml",
            subpackage="datareaders",
            classpackage=self.cfg.DEPLOYMENT.DATAREADER.DATAREADER,
        )
        data_reader_instance = data_reader()
        self.logger.info("Reading data with DataReader %s" % data_reader_instance.name)
        self.logger.info("Default CRAWLER is %s"%data_reader_instance.CRAWLER)
        self.logger.info("Default DATASET is %s"%data_reader_instance.DATASET)
        self.logger.info("Default GENERATOR is %s"%data_reader_instance.GENERATOR)
        # Update the generator...if needed
        if self._generatorClassQueueFlag:
            self.logger.info("Updating GENERATOR to queued class %s"%self._generatorClassQueue.__name__)
            data_reader_instance.GENERATOR = self._generatorClassQueue
            if self._generatorArgsQueueFlag:
                self.cfg.DEPLOYMENT.DATAREADER.GENERATOR_ARGS = (
                    self._generatorArgsQueue
                )
        else:
            if (
                self.cfg.DEPLOYMENT.DATAREADER.GENERATOR is not None
            ):
                self.logger.info("Updating GENERATOR using config specification to %s"%self.cfg.DEPLOYMENT.DATAREADER.GENERATOR)
                data_reader_instance.GENERATOR = locate_class(
                    package="ednaml",
                    subpackage="generators",
                    classpackage=self.cfg.DEPLOYMENT.DATAREADER.GENERATOR,
                )

        if self._crawlerClassQueueFlag: #here it checkes whether class flag is set, if it is then replace the build in class with custom class
            self.logger.info("Updating CRAWLER to %s"%self._crawlerClassQueue.__name__)
            data_reader_instance.CRAWLER = self._crawlerClassQueue
            if self._crawlerArgsQueueFlag: #check args also
                self.cfg.DEPLOYMENT.DATAREADER.CRAWLER_ARGS = (
                    self._crawlerArgsQueue
                )

        if self._crawlerInstanceQueueFlag:
            self.crawler = self._crawlerInstanceQueue
        else:
            self.crawler = self._buildCrawlerInstance(
                data_reader=data_reader_instance
            )
        
        # Only need test dataloader...
        self.buildTestDataloader(data_reader_instance, self.crawler)

    def buildTestDataloader(
        self, data_reader: DataReader, crawler_instance: Crawler
    ):
        """Builds a test dataloader instance given the data_reader class and a crawler instance that has been initialized

        Args:
            data_reader (DataReader): A datareader class
            crawler_instance (Crawler): A crawler instance
        """
        if self._testGeneratorInstanceQueueFlag:
            self.test_generator: Generator = self._testGeneratorInstanceQueue
        else:
            self.logger.info("Generating dataloader `{dataloader}` with `{mode}` mode".format(mode=self.dataloader_mode, dataloader=data_reader.GENERATOR.__name__))
            self.test_generator: Type[Generator] = data_reader.GENERATOR(
                logger=self.logger,
                gpus=self.gpus,
                transforms=self.cfg.TEST_TRANSFORMATION,
                mode=self.dataloader_mode,  # TODO convert this to better options: i.e. which mode to use, and which transformations to use, as an option in data_reader, specifically for deployments
                **self.cfg.DEPLOYMENT.DATAREADER.GENERATOR_ARGS
            )
            self.test_generator.build( 
                crawler_instance,
                batch_size=self.cfg.TEST_TRANSFORMATION.BATCH_SIZE,
                workers=self.cfg.TEST_TRANSFORMATION.WORKERS,
                **self.cfg.DEPLOYMENT.DATAREADER.DATASET_ARGS
            )

        if self.mode == "test":
            self.labelMetadata = self.test_generator.num_entities
        self.logger.info("Generated test data/query generator")

    def _buildCrawlerInstance(self, data_reader: DataReader) -> Crawler:
        """Builds a Crawler instance from the data_reader's provided crawler class in `data_reader.CRAWLER`

        Args:
            data_reader (DataReader): A DataReader class

        Returns:
            Crawler: A Crawler instanece for this experiment
        """
        return data_reader.CRAWLER(
            logger=self.logger, **self.cfg.DEPLOYMENT.DATAREADER.CRAWLER_ARGS
        )

    def deploy(self, **kwargs):
        self.deployment.deploy(continue_epoch=self.previous_stop + 1, **kwargs)

    def buildDeployment(self):
        """Builds the EdnaDeploy deployment and sets it up"""
        if self._deploymentClassQueueFlag:
            ExecutionDeployment = self._deploymentClassQueue
        else:
            ExecutionDeployment: Type[BaseDeploy] = locate_class(
                subpackage="deploy", classpackage=self.cfg.DEPLOYMENT.DEPLOY
            )
            self.logger.info(
                "Loaded {} from {} to build Deployment".format(
                    self.cfg.DEPLOYMENT.DEPLOY, "ednaml.deploy"
                )
            )

        if self._deploymentInstanceQueueFlag:
            self.deployment = self._deploymentInstanceQueue
        else:
            self.deployment = ExecutionDeployment(
                model=self.model,
                data_loader=self.test_generator.dataloader, # TODO
                epochs=self.cfg.DEPLOYMENT.EPOCHS,
                logger=self.logger,
                crawler=self.crawler,
                config=self.cfg,
                labels=self.labelMetadata,
                **self.cfg.DEPLOYMENT.DEPLOYMENT_ARGS
            )
            self.deployment.apply(
                step_verbose=self.cfg.LOGGING.STEP_VERBOSE,
                save_directory=self.saveMetadata.MODEL_SAVE_FOLDER,
                save_backup=self.cfg.SAVE.DRIVE_BACKUP,
                save_frequency=self.cfg.SAVE.SAVE_FREQUENCY,
                backup_directory=self.saveMetadata.CHECKPOINT_DIRECTORY,
                gpus=self.gpus,
                fp16=self.cfg.EXECUTION.FP16,
                model_save_name=self.saveMetadata.MODEL_SAVE_NAME,
                logger_file=self.saveMetadata.LOGGER_SAVE_NAME,
            )
    
    
    def resetDeploymentQueue(self):
        self._deploymentClassQueue = None
        self._deploymentClassQueueFlag = False
        self._deploymentInstanceQueue = None
        self._deploymentInstanceQueueFlag = False
    
    def addResetQueues(self):
        return [self.resetDeploymentQueue]

    def addDeploymentClass(self, deployment_class: Type[BaseDeploy]):
        self.logger.debug("Added deployment class: %s"%deployment_class.__name__)
        self._deploymentClassQueue = deployment_class
        self._deploymentClassQueueFlag = True

    def addDeployment(self, deployment: BaseDeploy):
        self._deploymentInstanceQueue = deployment
        self._deploymentInstanceQueueFlag = True


from torch import nn

from ednaml.utils import locate_class


class Loss(nn.Module):
    def __init__(self, lossname=None, metadata=None, **kwargs):
        super().__init__()

    def forward(self):
        raise NotImplementedError()


class TorchLoss(Loss):
    def __init__(self, lossname=None, metadata=None, **kwargs):
        super().__init__()
        self.lossclass = kwargs.get("loss_class")
        self.losskwargs = kwargs.get("loss_kwargs")
        lossclass = locate_class(
            package="torch", subpackage="nn", classpackage=self.lossclass
        )
        self.lossfn = lossclass(**self.losskwargs)

    def forward(
        self, input, target, **kwargs
    ):  # TODO this is hacky. need to generalize, or find a better way to represent the torch loss with its arguments...
        return self.lossfn(input, target, **kwargs)


from ednaml.loss.CenterLoss import CenterLoss
from ednaml.loss.ClusterLoss import ClusterLoss
from ednaml.loss.CompactContrastiveLoss import CompactContrastiveLoss
from ednaml.loss.ContrastiveLoss import ContrastiveLoss
from ednaml.loss.MarginLoss import MarginLoss
from ednaml.loss.ProxyNCA import ProxyNCA
from ednaml.loss.SoftmaxLabelSmooth import SoftmaxLabelSmooth
from ednaml.loss.SoftmaxLogitsLoss import SoftmaxLogitsLoss
from ednaml.loss.TripletLoss import TripletLoss



import torch
from torch import nn


def load_tf_weights_in_albert(model, config, tf_checkpoint_path):
    """Load tf checkpoints in a pytorch model."""
    import os
    import torch

    try:
        import re
        import numpy as np
        import tensorflow as tf
    except ImportError:
        raise ImportError(
            "Loading a TensorFlow model in PyTorch, requires TensorFlow to be"
            " installed. Please see https://www.tensorflow.org/install/ for"
            " installation instructions."
        )
    tf_path = os.path.abspath(tf_checkpoint_path)
    print("Converting TensorFlow checkpoint from {}".format(tf_path))
    if not os.path.exists(tf_path + "/checkpoint"):
        tf_path = tf_path + "/variables/variables"
    # Load weights from TF model
    init_vars = tf.train.list_variables(tf_path)
    names = []
    arrays = []
    for name, shape in init_vars:
        print("Loading TF weight {} with shape {}".format(name, shape))
        array = tf.train.load_variable(tf_path, name)
        names.append(name)
        arrays.append(array)
    for name, array in zip(names, arrays):
        name = name.replace("attention_1", "attention")
        name = name.replace("ffn_1", "ffn")
        name = name.split("/")
        # adam_v and adam_m are variables used in AdamWeightDecayOptimizer to calculated m and v
        # which are not required for using pretrained model
        if any(n in ["adam_v", "adam_m", "global_step"] for n in name):
            print("Skipping {}".format("/".join(name)))
            continue
        pointer = model
        for m_name in name:
            if re.fullmatch(r"[A-Za-z]+_\d+", m_name):
                l = re.split(r"_(\d+)", m_name)
            elif re.fullmatch(r"[A-Za-z]+_+[A-Za-z]+_\d+", m_name):
                l = re.split(r"_(\d+)", m_name)
            else:
                l = [m_name]
            if l[0] in ["LayerNorm", "attention", "ffn"] and len(l) >= 2:
                l = ["_".join(l[:-1])]
            if l[0] == "kernel" or l[0] == "gamma":
                pointer = getattr(pointer, "weight")
            elif l[0] == "output_bias" or l[0] == "beta":
                pointer = getattr(pointer, "bias")
            elif l[0] == "output_weights":
                pointer = getattr(pointer, "weight")
            elif l[0] == "squad":
                pointer = getattr(pointer, "classifier")
            else:
                try:
                    pointer = getattr(pointer, l[0])
                except AttributeError:
                    print("Skipping {}".format("/".join(name)))
                    continue
            if len(l) >= 2:
                num = int(l[1])
                pointer = pointer[num]

        if m_name[-11:] == "_embeddings":
            pointer = getattr(pointer, "weight")
        elif m_name == "kernel":
            array = np.transpose(array)
        try:
            assert pointer.shape == array.shape
        except AssertionError as e:
            e.args += (pointer.shape, array.shape)
            raise
        print("Initialize PyTorch weight {}".format(name))
        pointer.data = torch.from_numpy(array)
    return model


def prune_linear_layer(layer, index, dim=0):
    """Prune a linear layer (a model parameters) to keep only entries in index.
    Return the pruned layer as a new layer with requires_grad=True.
    Used to remove heads.
    """
    from torch import nn

    index = index.to(layer.weight.device)
    W = layer.weight.index_select(dim, index).clone().detach()
    if layer.bias is not None:
        if dim == 1:
            b = layer.bias.clone().detach()
        else:
            b = layer.bias[index].clone().detach()
    new_size = list(layer.weight.size())
    new_size[dim] = len(index)
    new_layer = nn.Linear(
        new_size[1], new_size[0], bias=layer.bias is not None
    ).to(layer.weight.device)
    new_layer.weight.requires_grad = False
    new_layer.weight.copy_(W.contiguous())
    new_layer.weight.requires_grad = True
    if layer.bias is not None:
        new_layer.bias.requires_grad = False
        new_layer.bias.copy_(b.contiguous())
        new_layer.bias.requires_grad = True
    return new_layer


def prune_conv1d_layer(layer, index, dim=1):
    """Prune a Conv1D layer (a model parameters) to keep only entries in index.
    A Conv1D work as a Linear layer (see e.g. BERT) but the weights are transposed.
    Return the pruned layer as a new layer with requires_grad=True.
    Used to remove heads.
    """
    from ednaml.utils.layers import Conv1D

    index = index.to(layer.weight.device)
    W = layer.weight.index_select(dim, index).clone().detach()
    if dim == 0:
        b = layer.bias.clone().detach()
    else:
        b = layer.bias[index].clone().detach()
    new_size = list(layer.weight.size())
    new_size[dim] = len(index)
    new_layer = Conv1D(new_size[1], new_size[0]).to(layer.weight.device)
    new_layer.weight.requires_grad = False
    new_layer.weight.copy_(W.contiguous())
    new_layer.weight.requires_grad = True
    new_layer.bias.requires_grad = False
    new_layer.bias.copy_(b.contiguous())
    new_layer.bias.requires_grad = True
    return new_layer


def prune_layer(layer, index, dim=None):
    """Prune a Conv1D or nn.Linear layer (a model parameters) to keep only entries in index.
    Return the pruned layer as a new layer with requires_grad=True.
    Used to remove heads.
    """
    from ednaml.utils.layers import Conv1D
    from torch import nn

    if isinstance(layer, nn.Linear):
        return prune_linear_layer(layer, index, dim=0 if dim is None else dim)
    elif isinstance(layer, Conv1D):
        return prune_conv1d_layer(layer, index, dim=1 if dim is None else dim)
    else:
        raise ValueError(
            "Can't prune layer of class {}".format(layer.__class__)
        )


class AlbertEmbeddingAverage(nn.Module):
    """Averages embeddings on given dimension

    Args:
        nn (_type_): _description_
    """

    def __init__(self) -> None:
        super().__init__()

    def forward(self, outputs):
        return torch.mean(outputs[0], dim=1)


class AlbertPooledOutput(nn.Module):
    """Torch layer to return a single index from list. It does not do anything"""

    def __init__(self):
        super().__init__()

    def forward(self, outputs):
        return outputs[1]


class AlbertRawCLSOutput(nn.Module):
    """Torch layer to return a single index from list. It does not do anything"""

    def __init__(self):
        super().__init__()

    def forward(self, outputs):
        return outputs[0][:, 0]



# BaseConfig class to manage the input configurations. TODO
import os
import json
from turtle import update
from typing import Dict, List
import warnings
import yaml
from ednaml.config import BaseConfig
from ednaml.config.DeploymentConfig import DeploymentConfig
from ednaml.config.ModelPluginConfig import ModelPluginConfig
from ednaml.utils import config_serializer
from ednaml.config.ConfigDefaults import ConfigDefaults

from ednaml.config.ExecutionConfig import ExecutionConfig
from ednaml.config.LoggingConfig import LoggingConfig
from ednaml.config.LossConfig import LossConfig
from ednaml.config.OptimizerConfig import OptimizerConfig
from ednaml.config.SaveConfig import SaveConfig
from ednaml.config.SchedulerConfig import SchedulerConfig
from ednaml.config.TransformationConfig import TransformationConfig
from ednaml.config.ModelConfig import ModelConfig
from ednaml.config.StorageConfig import StorageConfig
from ednaml.utils import merge_dictionary_on_key_with_copy

class EdnaMLConfig(BaseConfig):
    EXECUTION: ExecutionConfig
    DEPLOYMENT: DeploymentConfig
    SAVE: SaveConfig
    TRAIN_TRANSFORMATION: TransformationConfig
    TEST_TRANSFORMATION: TransformationConfig
    MODEL: ModelConfig
    LOSS: List[LossConfig]
    MODEL_PLUGIN: Dict[str, ModelPluginConfig]
    OPTIMIZER: List[
        OptimizerConfig
    ]  # one optimizer for each set of model params
    SCHEDULER: List[SchedulerConfig]  # one scheduler for each optimizer
    LOSS_OPTIMIZER: List[
        OptimizerConfig
    ]  # one optimizer for each loss with params
    LOSS_SCHEDULER: List[
        SchedulerConfig
    ]  # one scheduler for each loss_optimizer
    LOGGING: LoggingConfig
    STORAGE: StorageConfig

    extensions: List[str]

    def __init__(
        self, config_path: List[str], defaults: ConfigDefaults = ConfigDefaults(), **kwargs
    ):
        self.extensions = ["EXECUTION", "SAVE", "STORAGE", "TRANSFORMATION", "MODEL", "LOSS", "OPTIMIZER", "SCHEDULER", "LOSS_OPTIMIZER", "LOSS_SCHEDULER", "LOGGING", "DEPLOYMENT", "MODEL_PLUGIN"]  # TODO deal with other bits and pieces here!!!!!
        ydict = self.merge([self.read_path(config_component) for config_component in config_path])
        config_inject = kwargs.get("config_inject", None)
        if config_inject is not None and type(config_inject) is list:
            self.config_inject(ydict, config_inject)
        self._updateConfig(ydict, defaults, update_with_defaults=True)
    
    def config_inject(self, ydict, config_inject: List[List[str]]):
        for inject in config_inject:
            inject_key = inject[0]
            inject_value = inject[1]
            if len(inject) == 3:
                inject_kwarg = inject[2]
            else:
                inject_kwarg = None
            try:
                self._setinject(ydict, inject_key.split("."), inject_value, inject_kwarg)
                print("Injected key-value pair:  {key}, {value}".format(key=inject_key, value=inject_value))
            except KeyError:
                pass


    def _setinject(self, d, inject_key, inject_val, inject_kwarg = None):
        for elem in inject_key[:-1]:
            d = d[elem]
        if inject_kwarg is not None:
            d[inject_key[-1]][inject_kwarg] = inject_val    
        else:
            d[inject_key[-1]] = inject_val


    def _has_extension_verifier(self, ydict, extension, verification):
        return False if ydict.get(extension, verification) == verification else True

    def _updateConfig(self, ydict: Dict[str,str], defaults: ConfigDefaults, update_with_defaults: bool):
        added_extensions = []   # add control to check if extension was added or default was used...
        for extension in self.extensions:
            
            if extension == "EXECUTION":
                has_extension = self._has_extension_verifier(ydict, extension, {})
                if has_extension or update_with_defaults:
                    self.EXECUTION = ExecutionConfig(ydict.get(extension, {}), defaults) 
                    added_extensions.append([extension])
            elif extension == "DEPLOYMENT":
                has_extension = self._has_extension_verifier(ydict, extension, {})
                if has_extension or update_with_defaults:
                    self.DEPLOYMENT = DeploymentConfig(ydict.get(extension, {}), defaults)
                    added_extensions.append([extension])
            elif extension == "STORAGE":
                has_extension = self._has_extension_verifier(ydict, extension, {})
                if has_extension or update_with_defaults:
                    self.STORAGE = StorageConfig(ydict.get(extension, {}), defaults)
                    added_extensions.append([extension])
            elif extension == "SAVE":
                has_extension = self._has_extension_verifier(ydict, extension, {})
                if has_extension or update_with_defaults:
                    self.SAVE = SaveConfig(ydict.get(extension, {}), defaults)
                    added_extensions.append([extension])
            elif extension == "TRANSFORMATION":
                has_extension = self._has_extension_verifier(ydict, extension, {})
                has_train = self._has_extension_verifier(ydict, "TRAIN_TRANSFORMATION", {})
                if (has_extension and has_train) or update_with_defaults:
                    self.TRAIN_TRANSFORMATION = TransformationConfig(
                        dict(
                            merge_dictionary_on_key_with_copy(ydict.get(extension, {}), ydict.get("TRAIN_TRANSFORMATION", {}))
                        ),
                        defaults,
                    )
                    added_extensions.append(["TRAIN_TRANSFORMATION"])
                has_test = self._has_extension_verifier(ydict, "TEST_TRANSFORMATION", {})
                if (has_extension and has_test) or update_with_defaults:
                    self.TEST_TRANSFORMATION = TransformationConfig(
                        dict(
                            merge_dictionary_on_key_with_copy(ydict.get(extension, {}), ydict.get("TEST_TRANSFORMATION", {}))
                        ),
                        defaults,
                    )
                    added_extensions.append(["TEST_TRANSFORMATION"])
            elif extension == "MODEL":
                has_extension = self._has_extension_verifier(ydict, extension, {})
                if has_extension or update_with_defaults:
                    self.MODEL = ModelConfig(
                        ydict.get(extension, {}), defaults
                    )  # No default MODEL itself, though it will be instantiated here? deal with this TODO
                    added_extensions.append([extension])
            elif extension == "MODEL_PLUGIN":
                has_extension = self._has_extension_verifier(ydict, extension, [])
                if has_extension or update_with_defaults:
                    mp_list = [ModelPluginConfig(plugin_item, defaults) for plugin_item in ydict.get(extension, [])]
                    self.MODEL_PLUGIN = {item.PLUGIN_NAME: item for item in mp_list}
                    added_extensions.append([extension])
            elif extension == "LOSS":
                has_extension = self._has_extension_verifier(ydict, extension, [])
                if has_extension or update_with_defaults:
                    self.LOSS = [
                        LossConfig(loss_item, defaults)
                        for loss_item in ydict.get(extension, [])
                    ]  # No default LOSS itself --> it will be empty...
                    added_extensions.append([extension])
            elif extension == "OPTIMIZER":
                # Default optimizer is Adam
                has_extension = self._has_extension_verifier(ydict, extension, [{}])
                if has_extension or update_with_defaults:
                    self.OPTIMIZER = [
                        OptimizerConfig(optimizer_item, defaults)
                        for optimizer_item in ydict.get(extension, [{}])
                    ]
                    added_extensions.append([extension])
            elif extension == "SCHEDULER":
                has_extension = self._has_extension_verifier(ydict, extension, [{}])
                if has_extension or update_with_defaults:
                    self.SCHEDULER = [
                        SchedulerConfig(scheduler_item, defaults)
                        for scheduler_item in ydict.get(extension, [{}])
                    ]
                    added_extensions.append([extension])
            elif extension == "LOSS_OPTIMIZER":
                has_extension = self._has_extension_verifier(ydict, extension, [{}])
                if has_extension or update_with_defaults:
                    self.LOSS_OPTIMIZER = [
                        OptimizerConfig(optimizer_item, defaults)
                        for optimizer_item in ydict.get(extension, [{}])
                    ]
                    added_extensions.append([extension])
            elif extension == "LOSS_SCHEDULER":
                has_extension = self._has_extension_verifier(ydict, extension, [{}])
                if has_extension or update_with_defaults:
                    self.LOSS_SCHEDULER = [
                        SchedulerConfig(scheduler_item, defaults)
                        for scheduler_item in ydict.get(extension, [{}])
                    ]
                    added_extensions.append([extension])
            elif extension == "LOGGING":
                has_extension = self._has_extension_verifier(ydict, extension, {})
                if has_extension or update_with_defaults:
                    self.LOGGING = LoggingConfig(ydict.get(extension, {}), defaults)
                    added_extensions.append([extension])
            else:
                raise RuntimeError("Somehow received unhandled extension %s"%str(extension))
        return added_extensions
        # Things without defaults that MUST be provided: model ✅, train_dataloader, loss ✅, trainer TODO

    def export(self, mode="yaml"):
        dicts = json.dumps(self, default=config_serializer)  # or getvars()????
        dicta = json.loads(dicts)
        if mode == "yaml":
            return yaml.dump(dicta)
        elif mode == "json":
            return dicts
        elif mode == "dict":
            return dicta

    def extend(self, config_path: str, defaults: ConfigDefaults = ConfigDefaults(), **kwargs):
        """Extends the existing config object with fields from the provided config

        Args:
            config_path (str): The path to the config file to extend with
        """        
        responses = {}
        if type(config_path) is str:
            config_path = [config_path]
        if type(config_path) is list:
            for cpath in config_path:
                responses[cpath] = self._extend(cpath, defaults, **kwargs)
        else:
            raise ValueError("Expected `list`, got %s"%type(config_path))

        return responses

    def _extend(self, config_path, defaults: ConfigDefaults, **kwargs):
        ydict = self.read_path(config_path)
        config_inject = kwargs.get("config_inject", None)
        if config_inject is not None and type(config_inject) is list:
            self.config_inject(ydict, config_inject)
        added_extensions = self._updateConfig(ydict, defaults, update_with_defaults=False)
        return "Extended with : %s"%", ".join([item[0] for item in added_extensions if len(item)>0])



    def read_path(self, path: os.PathLike) -> Dict[str,str]:
        """Reads the config at the path and yields a dictionary of the config

        Args:
            path (os.PathLike): Path to the config

        Raises:
            FileNotFoundError: Raised if path does not exist

        Returns:
            Dict[str,str]: Config as a python dictionary
        """        
        ydict = {} 

        if len(path) > 0:
            if not os.path.exists(path):
                raise FileNotFoundError(
                    "No file found for config at : %s" % path
                )
            else:
                with open(path, "r") as cfile:
                    ydict = yaml.safe_load(cfile.read().strip()) #loading the configuration file.
        return ydict

    def _merge(self,base_config: Dict[str,str], extension: Dict[str,str], recursion_flag: bool = False):
        """Combines two config files
        If capital lettered key is present it'll be replaced
        If small lettered key is present, it'll be deleted and considered the key in second object only

        Args:
            first object: Dict (base config)
            second object: Dict (second config)
        Returns:
            Dict[str,str]: Config as a combined python dictionary
        """   
        #if(recursion_flag):
        #    to_keep = lambda key: key.islower()
        #    {k: v for k, v in base_config.items() if k not in keyfilter(to_keep, base_config)}
        lower_flag = False
        upper_flag = False
        for extension_key in extension.keys():
            if extension_key in base_config:
                if extension_key.islower(): # If NOT a built-in key, we will do a full replacement of the parent...
                    lower_flag = True
                else:   # If it is a built-in key
                    upper_flag = True
                    if isinstance(base_config[extension_key], dict):    # If it is a dict
                        if isinstance(extension[extension_key], dict):
                            # Recurse if key is a Dictionary with elements inside.
                            base_config[extension_key] = self._merge(base_config[extension_key], extension[extension_key],recursion_flag=True)
                        else:
                            base_config[extension_key]=extension[extension_key]
                            warnings.warn("Key {key}: Dict key in base configuration is not a dictionary in replacement extension. \n\tBase: {base_str}\n\tExtension: {ext_str}".format(
                                key=extension_key,
                                base_str = str(base_config(extension_key)),
                                ext_str = str(base_config(extension_key)),
                            ))
                    else:   # The base config key is not a dictionary...we just do replacement. Don't care if replacement is dictionary
                        base_config[extension_key]=extension[extension_key] 
            else:   # a key in extensionis NOT in base_config, so we don't really care...
                base_config[extension_key] = extension[extension_key]
        if lower_flag and upper_flag:
            raise ValueError("built-in and custom keys cannot mix")
        if lower_flag:
            return extension
        return base_config

    def merge(self, config_paths: List[str]):
        """Combines multiple config files
        Args:
            first object: Dict array (base config as index 0)
        Returns:
            Dict[str,str]: Config as a combined python dictionary
        """  
        base_config = config_paths[0]
        for extension_config in config_paths[1:]:
            base_config = self._merge(base_config, extension_config)
        return base_config



"""
Notes for LOSS_OPTIMIZER and LOSS_SCHEDULER
        Now we import the loss-optimizers and loss-scheduler
        Here, we need some additional code to make things work.
        
        Situation 1. LossBuilders, plus fully enumerated loss-optimizers. Simply load them into the config
        Situation 2. LossBuilders with no learnable params, and no loss-optimizers.
        Situation 3. LossBuilders with learnable parameters, and no loss-optimizers.
        Situation 4. LossBuilders, some with learnable parameters, and only corresponding loss-optimizers
        Situation 5. LossBuilders, some with learnable parameters, and less loss-optimizers
        Situation 6. LossBuilders, some with learnable parameters, and more loss-optimizers

        To cover all cases elegantly, we will simply provide:
        Situation 1. All loss-optimizers
        Situation 2. No way to check params here. Provide a default loss-optimizer
        Situation 3. No way to check params here. Provide a default loss-optimizer
        Situation 4. Provide enumerated loss-optimizers
        Situation 5. Provide enumerated loss-optimizers
        Situation 6. Provide enumerated loss-optimizers


        Then, EdnaML will:
        - construct a temporary dictionary of loss-optimizer configs, ordered by their name
        - For each loss-builder, check if there is a loss-optimizer with that name in the dict
        - If there is, then construct optimizer with those parameters
        - If there is not, then use the first optimizer's parameters to construct missing ones
        - this way, for each situation:
            Situation 1. No problems
            Situation 2. Default parameters used, but the optimizer builder will return None
            Situation 3. Default parameters used.
            Situation 4. No problems. The LossBuilders without params will return None
            Situation 5. Missing optimizers will be replaced with the first optimizer
            Situation 6. Extra optimizers will do nothing

        For Loss-schedulers, either there are enumerated schedulers, or a single default
        in EdnaML, we will match schedulers to the corresponding optimizers
        For optimizer without scheduler, create a scheduler using the first one
"""



from urllib.parse import urlparse
import sys, requests
import sys
import json
import os
import shutil
import tempfile
import fnmatch
from functools import wraps
from hashlib import sha256
from io import open
from tqdm import tqdm

try:
    import boto3
    from botocore.config import Config as botoConfig
    from botocore.exceptions import ClientError as botoClientError
except ImportError:
    boto3 = None
    botoConfig = None
    botoClientError = None

# from https://sumit-ghosh.com/articles/python-download-progress-bar/
def download(file_, url):
    sys.stdout.write("Downloading %s from %s" % (file_, url))
    with open(file_, "wb") as f:
        response = requests.get(url, stream=True)
        all_data = response.headers.get("content-length")
        if all_data is None:
            f.write(response.content)
        else:
            all_data = int(all_data)
            progress = 0
            for chunk in response.iter_content(
                chunk_size=max(int(all_data / 1000), 1024 * 1024)
            ):
                progress += len(chunk)
                f.write(chunk)
                sys.stdout.write(
                    "\r{}/{} bytes [{}{}]".format(
                        all_data,
                        progress,
                        "█" * int(100 * progress / all_data),
                        "." * (100 - int(100 * progress / all_data)),
                    )
                )
                sys.stdout.flush()
    sys.stdout.write("\nDownload of %s to %s completed\n" % (file_, url))


def cached_path(
    url_or_filename, cache_dir=None, force_download=False, proxies=None
):
    """
    Given something that might be a URL (or might be a local path),
    determine which. If it's a URL, download the file and cache it, and
    return the path to the cached file. If it's already a local path,
    make sure the file exists and then return the path.
    Args:
        cache_dir: specify a cache directory to save the file to (overwrite the default cache dir).
        force_download: if True, re-dowload the file even if it's already cached in the cache dir.
    """
    import os
    from torch.hub import _get_torch_home
    from pathlib import Path

    if cache_dir is None:
        cache_dir = os.getenv(
            "PYTORCH_TRANSFORMERS_CACHE",
            os.getenv(
                "PYTORCH_PRETRAINED_BERT_CACHE",
                os.path.join(_get_torch_home(), "pytorch_transformers"),
            ),
        )
    if sys.version_info[0] == 3 and isinstance(url_or_filename, Path):
        url_or_filename = str(url_or_filename)
    if sys.version_info[0] == 3 and isinstance(cache_dir, Path):
        cache_dir = str(cache_dir)

    parsed = urlparse(url_or_filename)

    if parsed.scheme in ("http", "https", "s3"):
        # URL, so get it from the cache (downloading if necessary)
        return get_from_cache(
            url_or_filename,
            cache_dir=cache_dir,
            force_download=force_download,
            proxies=proxies,
        )
    elif os.path.exists(url_or_filename):
        # File, and it exists.
        return url_or_filename
    elif parsed.scheme == "":
        # File, but it doesn't exist.
        raise EnvironmentError("file {} not found".format(url_or_filename))
    else:
        # Something unknown
        raise ValueError(
            "unable to parse {} as a URL or as a local path".format(
                url_or_filename
            )
        )


def get_from_cache(url, cache_dir=None, force_download=False, proxies=None):
    """
    Given a URL, look for the corresponding dataset in the local cache.
    If it's not there, download it. Then return the path to the cached file.
    """
    import os
    from torch.hub import _get_torch_home
    from pathlib import Path

    if cache_dir is None:
        cache_dir = os.getenv(
            "PYTORCH_TRANSFORMERS_CACHE",
            os.getenv(
                "PYTORCH_PRETRAINED_BERT_CACHE",
                os.path.join(_get_torch_home(), "pytorch_transformers"),
            ),
        )
    if sys.version_info[0] == 3 and isinstance(cache_dir, Path):
        cache_dir = str(cache_dir)
    if sys.version_info[0] == 2 and not isinstance(cache_dir, str):
        cache_dir = str(cache_dir)

    if not os.path.exists(cache_dir):
        os.makedirs(cache_dir)

    # Get eTag to add to filename, if it exists.
    if url.startswith("s3://"):
        etag = s3_etag(url, proxies=proxies)
    else:
        try:
            response = requests.head(url, allow_redirects=True, proxies=proxies)
            if response.status_code != 200:
                etag = None
            else:
                etag = response.headers.get("ETag")
        except EnvironmentError:
            etag = None

    if sys.version_info[0] == 2 and etag is not None:
        etag = etag.decode("utf-8")
    filename = url_to_filename(url, etag)

    # get cache path to put the file
    cache_path = os.path.join(cache_dir, filename)

    # If we don't have a connection (etag is None) and can't identify the file
    # try to get the last downloaded one
    if not os.path.exists(cache_path) and etag is None:
        matching_files = fnmatch.filter(os.listdir(cache_dir), filename + ".*")
        matching_files = list(
            filter(lambda s: not s.endswith(".json"), matching_files)
        )
        if matching_files:
            cache_path = os.path.join(cache_dir, matching_files[-1])

    if not os.path.exists(cache_path) or force_download:
        # Download to temporary file, then copy to cache dir once finished.
        # Otherwise you get corrupt cache entries if the download gets interrupted.
        with tempfile.NamedTemporaryFile() as temp_file:
            print(
                "%s not found in cache or force_download set to True,"
                " downloading to %s",
                url,
                temp_file.name,
            )

            # GET file object
            if url.startswith("s3://"):
                s3_get(url, temp_file, proxies=proxies)
            else:
                http_get(url, temp_file, proxies=proxies)

            # we are copying the file before closing it, so flush to avoid truncation
            temp_file.flush()
            # shutil.copyfileobj() starts at the current position, so go to the start
            temp_file.seek(0)

            print("copying %s to cache at %s", temp_file.name, cache_path)
            with open(cache_path, "wb") as cache_file:
                shutil.copyfileobj(temp_file, cache_file)

            print("creating metadata file for %s", cache_path)
            meta = {"url": url, "etag": etag}
            meta_path = cache_path + ".json"
            with open(meta_path, "w") as meta_file:
                output_string = json.dumps(meta)
                if sys.version_info[0] == 2 and isinstance(output_string, str):
                    output_string = unicode(
                        output_string, "utf-8"
                    )  # The beauty of python 2
                meta_file.write(output_string)

            print("removing temp file %s", temp_file.name)

    return cache_path


def s3_request(func):
    """
    Wrapper function for s3 requests in order to create more helpful error
    messages.
    """

    @wraps(func)
    def wrapper(url, *args, **kwargs):
        try:
            return func(url, *args, **kwargs)
        except botoClientError as exc:
            if int(exc.response["Error"]["Code"]) == 404:
                raise EnvironmentError("file {} not found".format(url))
            else:
                raise

    return wrapper


@s3_request
def s3_etag(url, proxies=None):
    """Check ETag on S3 object."""
    s3_resource = boto3.resource("s3", config=botoConfig(proxies=proxies))
    bucket_name, s3_path = split_s3_path(url)
    s3_object = s3_resource.Object(bucket_name, s3_path)
    return s3_object.e_tag


@s3_request
def s3_get(url, temp_file, proxies=None):
    """Pull a file directly from S3."""
    s3_resource = boto3.resource("s3", config=botoConfig(proxies=proxies))
    bucket_name, s3_path = split_s3_path(url)
    s3_resource.Bucket(bucket_name).download_fileobj(s3_path, temp_file)


def http_get(url, temp_file, proxies=None):
    req = requests.get(url, stream=True, proxies=proxies)
    content_length = req.headers.get("Content-Length")
    total = int(content_length) if content_length is not None else None
    progress = tqdm(unit="B", total=total)
    for chunk in req.iter_content(chunk_size=1024):
        if chunk:  # filter out keep-alive new chunks
            progress.update(len(chunk))
            temp_file.write(chunk)
    progress.close()


def url_to_filename(url, etag=None):
    """
    Convert `url` into a hashed filename in a repeatable way.
    If `etag` is specified, append its hash to the url's, delimited
    by a period.
    """
    url_bytes = url.encode("utf-8")
    url_hash = sha256(url_bytes)
    filename = url_hash.hexdigest()

    if etag:
        etag_bytes = etag.encode("utf-8")
        etag_hash = sha256(etag_bytes)
        filename += "." + etag_hash.hexdigest()

    return filename


def split_s3_path(url):
    """Split a full s3 path into the bucket name and path."""
    parsed = urlparse(url)
    if not parsed.netloc or not parsed.path:
        raise ValueError("bad s3 path {}".format(url))
    bucket_name = parsed.netloc
    s3_path = parsed.path
    # Remove '/' at beginning of path.
    if s3_path.startswith("/"):
        s3_path = s3_path[1:]
    return bucket_name, s3_path



"""
ModelConversionDeployment is a built-in Edna class to convert models between different ModelAbstract implementations.

The use case for such a ModelConversionDeployment is to change how a model functions after it has already been trained, such as adding additional outputs like moden confidence or self-diagnoses, adjusting how a model manipulates its input, including a post-facto reject option, etc.

Prerequisites for using ModelConversionDeployment:

- a trained model's saved weights

Corequisites for using ModelConversionDeployment:

- a model class, inherited from ModelAbstract
    - ModelAbstract already contains methods to load weights from a .pth file where the keys do not exactly match, with partial_load. This would normally be used to load the saved weights of the older model into the new model
- a config file that links to the old weights through the Edna section OR the save section
- NOTE --> we still need to deal with one small tiny issue!!!!!!! when providing multiple configs, the latter replace the older on same keys. So will the deployment.yml's SAVE section replace the model.yml's SAVE section????
- a config file with an updated Save section to push the new model and weights
- any helper files -- if not provided, ModelConversionDeployment will use the classes from the original models' implementation, unless a new Crawler is provided
- 


Usage Instructions:


edna ModelConversionDeployment new_model_file.py model_config.yml deployment_config.yml

"""

from ednaml.deploy import BaseDeploy
# So we need the load() to be partial load, possibly...
# WE NEED TO ADJUST HOW MODEL LOADS ARE HANDLED IN EDNAML AND EDNADEPLOY

class ModelConversionDeployment(BaseDeploy):
  def deploy_step(self, batch):
    batch = tuple(item.cuda() for item in batch)
    all_input_ids, all_attention_mask, all_token_type_ids, all_masklm, all_labels = batch
    # outputs is a tuple of output, features, secondary_output --> which has the plugin components. Maybe we will split plugin outputs separately...
    outputs = self.model(all_input_ids, token_type_ids = all_token_type_ids, attention_mask=all_attention_mask)
    logit = outputs[0].detach().cpu()
    logit_labels = torch.argmax(logit, dim=1)
    #print(str(logit_labels[0]), "\t", ",".join([str(item) for item in all_input_ids[0].cpu()[:10].tolist()]))
    
    # What we need to return...
    # - predicted label, raw logit probability, model L-score, L-score perturbation, Model L-score, cosine distance to nearest proxy in model, euclidean distance to nearest proxy in model
    # So, we need the models themselves to return some of this stuff...
    # Models have already been trained. We basically want the model to be equipped with some plugins
    return logit_labels, logit, None

  def output_setup(self, **kwargs):
    file_path = kwargs.get("file_name", "deploy-out.txt")
    self.fobj = open(file_path, "w")

  def output_step(self, logits, features, secondary):
    self.fobj.write("\n".join([str(item) for item in logits.tolist()])+"\n")